# Rudy
Deep Learning for Radar data.

## Data
For training the deep network we're using Range Doppler matrices, which looks similary to images and share with them some properties.

### Range Doppler Matrices (RD Matrices)
RD Matrices are calculated from raw signal using 2D FFT. For every radar sensor we got 4 matrices in total - *amplitude* and *phase* matrix for each of 2 channels. 

## Using scripts

### run_tasks_dnn.py, run_tasks_mac.py
All tasks to run all experiments all already specifed in the file `run_tasks_dnn.py` or file `run_tasks_mac.py`, where I've tested all functionality.

To give a overview of specific tasks


### Tasks (example: run_tasks_mac.py)
In this script you can find Main method, where are defined all tasks with actions, which needs to be performed sequentially. You can add your own task, just by creating the instance of Task:

*1*. First of all you need to specify the configuration for the task, it can be **global** or **task-specific**, we defined it globaly.
```python
config = {
  # What shape the input data have, from the left it is (batch number, time = need to be set to 1, and image size (16, 64, 3))
  'batch_input_shape' : (batch_size, 1) + img_size,
  # Number of epochs the neural net will continue to train
  'epochs' : 100,
  # Path to data or folder __h5__ where the cached is stored.
  'cache_path' : '/media/tesla/PRA1-HDD004',
  # Path where you want to save the experiments (predictions and weights)
  'experiments_path' : '/media/tesla/PRA1-HDD004/Experiments',
  # Moni server address
  'moni_server' : 'http://localhost:9000',
  # If you want to plot training progress
  'plotter' : False,
  # What optimizer you want to run ('adamax', 'sgd')
  'optimizer' : 'adamax',
  # How you want to reset the LSTM ('auto' means at the end of each dataset)
  'lstm_reset' : 'auto',
  # Save weights for each epoch
  'train_checkpoint' : True,
  # Stop early if the neural net will stop learn
  'train_early_stop' : False,
  # Filename for exported weights
  'weights_file' : 'cnn_lstm',
  # Learning rate
  'learn_rate' : 0.1,
  # Fit using generator or load the date into memory (memory intensive, but faster training and you can use shuffle in SimpleForward model)
  'fit_using_generator' : False,
  # Weight for specific classes, 0 ... no alert, 1 ... alert
  'class_weights' : {0: 1, 1: 100},
  # Type of recurrent network you want to use ('lstm', 'gru', 'simple_rnn')
  'rnn_class' : 'lstm',
}
```
*2*. As next we can create as many tasks as we want.
```python
task_train_multids =  Task(config,
                        path = 'Func_Test_OTRs',
                        actions = [
                                CreateSimple,
                                CompileModel,
                                Comment('Model created.'),
                                LoadDatasets(datasets),
                                Comment('Multidataset was loaded.'),
                                TrainModel,
                                Comment('Model trained.'),
                                BenchmarkModel,
                                Comment('Benchmark of model is done.'),
                            ]),
```

*3*. If you want, you can create your own actions:
```python
def MyCustomAction(task, error):
  '''
  Action is just a simple function that takes two arguments:
  task - reference to the task object in which the action is performed
  error - reference to the last error (from previous action) or None, if there wasn't any previous action
  Action has to return the an Error object, but you can return NoError, if there is no errors.
  '''
  
  # You can use any attribute or method inside task object
  # For example, you can just print name and path of the task
  print(task.name, task.path)
  
  # Don't forget to return NoError if everything was okay.
  return NoError()
```
*4*. You can subclass `ActionError` to create your own custom error type, or subclass `Task` or `BaseTask` to create more sophisticated tasks.
```python
class AwesomeTask(Task):
  # You can really easily add more attributes to the task
  def __init__(self, config, path, actions = [], files = []):
    super().__init__(config, path, actions)
    self.files = files
  
class NoElementsError(ActionError):
  pass
    
# Now we can use it in the action
def AwesomeAction(task, error):
  if files:
    for file in files:
      path = os.path.join(path, file)
      print(path)
    return NoError()
  else:
    return NoElementsError('There are no elements in the task.files attribute.')
```
*5*. Now you can use the new custom **tasks** and **actions**.
```python
awesome_task = AwesomeTaks(config,
  path = 'Test', # Specify a path
  actions = [
    MyCustomAction,
    AwesomeAction,
  ],
  files = [
    'file0.csv',
    'file1.csv',
    'file2.csv',
  ])
```
*6*. To run the tasks, just add them to list or dictionary and pass it to `RunTasks` function. If you create dictionary with tasks, you can then call the specific task from command line: `run_tasks.py train_multids`. Dictionary is always preferable way to specify tasks, you can also run `run_tasks.py ls` to show a list of all availible tasks. 
```python
tasks = [
  task_train_multids,
  awesome_task
]
# OR
tasks = {
  'train_multids' : task_train_multids,
  'awesome_task' : awesome_task,
}

RunTasks(tasks)
```

### CSVConvert.py
Usage: 
```
CSVConvert.py [-h] [-o OUTPUT] [-v] [-s] [-l LOG] [--version] PATH
```
CSVConvert is responsible for extracting the values from CSV files and calculate all the RD Matrices. All Frames contaninig RD Matrices will be saved in a `rudy.HStorage` class which can be serialized (cached) using **h5py** (by default it is cached).

### CreateDatasets.py
Usage: 
```
CreateDatasets.py [-h] [-t TRAIN_SIZE] [-s SENSOR_SIDE] [-b BEAM]
                  [-p POSITIVE_SKIP] [-n NEGATIVE_SKIP] [-r RATIO] [-f]
                  [-v] [-l LOG] [--version]
                  PATH
```
CreateDatasets as the name suggest is creating datasets from calculated RD Matrices. Each script is independed, if the CreateDatasets script is run and the cache of `rudy.HStorage` isn't availible, it will be created from CSV files. Final datasets is cached again using **h5py**.

### dnn.py
Usage:
```
dnn.py [-h] -i INPUT [-e EPOCH] [-b BATCH]
```
Script that build the DNN model using [Keras](https://github.com/fchollet/keras) with [Theano](https://github.com/Theano/Theano) backend. If the cache version of datasets isn't found, it will be created from `rudy.Storage` if it's availible, otherwise it will use CSV files to calculate everything from scratch (cache is turn on by default, so next run should be faster.