import rudy.utils_logging as log
from rudy.data_sets import DataSets, DataSet
from rudy.utils_tasks import NoError

from dnn_tasks import FindBasepath, CreateFilepathIfNecessary

from datetime import datetime
import numpy as np
import pickle
import json
import os

from colorama import Fore, Back, Style

def PlotPredictions(y_pred, y_orig, save_to_file = '', file_exts = ['pdf'], show_plot = False):
	import matplotlib.pyplot as plt
	plt.ion()

	y_pred = np.array(y_pred).flatten()
	y_orig = np.array(y_orig).flatten()

	def Plot(y_pred, y_orig, x = []):
		if len(x) == 0:
			x = np.arange(len(y_pred))

		plt.figure(figsize = (8.0, 6.0))
		# Draw Prdecitions
		plt.subplot(3,1,1)
		# plt.title('Predictions', fontsize=20)
		plt.fill_between(x, y_pred, color = 'green', alpha = 0.4)
		plt.plot(x, y_pred, color = 'green', linewidth = 0.5)
		plt.ylim(0, 1)
		# plt.xlabel('xlabel', fontsize=18)
		plt.ylabel('Predicted alert', fontsize=16)

		# Draw Originals
		plt.subplot(3,1,2)
		# plt.title('Original', fontsize=20)
		plt.fill_between(x, y_orig, color = 'blue', alpha = 0.4)
		plt.plot(x, y_orig, color = 'blue', linewidth = 0.5)
		plt.ylim(0, 1)
		# plt.xlabel('xlabel', fontsize=18)
		plt.ylabel('Original alert', fontsize=16)

		# Draw Differences
		plt.subplot(3,1,3)
		# plt.title('Diff', fontsize=20)
		d = y_orig - y_pred
		plt.fill_between(x, d, color = 'red', alpha = 0.4)
		plt.plot(x, d, color = 'red', linewidth = 0.5)
		plt.plot(x, [0] * len(x), color = 'black', linestyle = '--')
		plt.ylim(-1, 1)
		plt.xlabel('MCC', fontsize=16)
		plt.ylabel('Error', fontsize=16)

		# Show plot
		if show_plot:
			plt.show()

	# Plot
	Plot(y_pred, y_orig)

	# Save
	if save_to_file:
		for ext in file_exts:
			plt.savefig('{name}.{ext}'.format(name = save_to_file, ext = ext), dpi = 100)

	if not show_plot:
		plt.close()

	# Plot detail
	d = np.abs(y_orig - y_pred)
	tmp = d[y_orig == 0]
	d[y_orig == 0] = 100
	im = np.argmin(d)
	# im = np.argmax(np.abs(y_orig - y_pred))
	x = range(max(im - 100, 0), min(im + 100, len(y_pred)))
	Plot(y_pred[x], y_orig[x], np.array(x))

	# Save detail
	if save_to_file:
		for ext in file_exts:
			plt.savefig('{name}_detail.{ext}'.format(name = save_to_file, ext = ext), dpi = 100)

	if not show_plot:
		plt.close()

def GetAllPredictions(model, storage, batch_input_shape = (1, 16, 64, 3), root = '', choosen_side = 'REAR_LEFT', calculate_stats = True, save_to_file = False, plot_comparison = False, model_id = ''):
	for batch in storage.batches:
		ds = DataSets(path = storage._path, batch_name = batch.name, choosen_side = choosen_side, fixed_beams = [2], cache_path = storage._cache_path, batch_input_shape = batch_input_shape)
		if ds.cached and len(ds.all) != 0:
			log.info('Benchmarking dataset \'{name}\' of size {size}:'.format(name = batch.name, size = len(ds.all)))
			y_orig = ds.all.labels
			x_orig = ds.all.frames
			y_pred = model.predict(x_orig, batch_size = batch_input_shape[0], verbose = 1)

			score = 0
			accuracy = 0

			if calculate_stats:
				score, accuracy = model.evaluate(x_orig, y_orig, batch_size = batch_input_shape[0], verbose = 1)
				log.info('Score: {}'.format(score))
				log.info('Accuracy: {}'.format(accuracy))

			if save_to_file:
				base_dir = FindBasepath(root = root, subfolder = 'Benchmarks.{}'.format(model_id), use_timefolder = True, find_last = True)
				name = '{name}.pickle'.format(name = batch.name)
				with open(CreateFilepathIfNecessary(os.path.join(base_dir, name)), 'wb') as f:
					pickle.dump({
						'name' 		: batch.name,
						'y_orig' 	: y_orig,
						'y_pred' 	: y_pred,
						'score' 	: score,
						'accuracy' 	: accuracy,
						}, f, pickle.HIGHEST_PROTOCOL)
				log.debug('-' * 16 + ' File saved ' + '-' * 16)

			if plot_comparison:
				PlotPredictions(y_pred, y_orig, save_to_file = os.path.join(root, name), file_exts = ['png', 'pdf'])
				PlotROC(y_pred, y_orig, save_to_file = file_name, file_exts = ['png', 'pdf'])
		else:
			log.warning('Batch {name} has no {side} data.'.format(name = batch.name, side = choosen_side))

def CalculateTPTN(y_pred, y_orig, threshold):
	'''Function that calculates all error rates for given threshold.'''
	num_pos = np.sum(y_pred > threshold)
	num_neg = len(y_pred) - num_pos
	y  = (y_pred > threshold).astype(int)
	# This way we can distinguish between all stats with one equation
	ys = y * 2 + y_orig
	# ys == 0 ... true_negative
	# ys == 1 ... false_negative
	# ys == 2 ... false_positive
	# ys == 3 ... true_positive
	true_negative  = np.sum(ys == 0)
	false_negative = np.sum(ys == 1)
	false_positive = np.sum(ys == 2)
	true_positive  = np.sum(ys == 3)

	true_positive_rate = true_positive / (true_positive + false_negative)
	false_positive_rate = false_positive / (false_positive + true_negative)

	return false_positive_rate, true_positive_rate

def PlotROC(y_pred, y_orig, num = 200, show_plot = False, save_to_file = '', file_exts = ['pdf'], automatic_focus = False):
	'''Plot ROC curve with given number of samples, num = 100 means spacing between samples = 1/100 = 0.01.
	If automatic_focus is True, function will try to focus on places in ROC with high 1st order difference.
	'''
	accuracy = 1.0 / num
	thresholds = np.linspace(0.0, 1.0, num = num)
	roc = np.matrix([CalculateTPTN(y_pred, y_orig, th) for th in thresholds])

	if automatic_focus:
		# Get 1st order differences of True positives value in ROC
		d = np.abs(np.diff(np.array(roc[:,0]).flatten()))
		focus_points = thresholds[:-1][d > accuracy] # positions that needs to be focused
		# focus_ranges = np.array([(max(x - accuracy, 0.0), min(1.0, x + accuracy)) for x in focus_points])

		focus_ranges = [0.0]
		for point in focus_points:
			start_point = max(0.0, point - accuracy)
			end_point   = min(1.0, point + accuracy)

			if focus_ranges[-1] > start_point:
				focus_ranges[-1] = end_point
			else:
				if start_point != 0.0:
					focus_ranges.append(start_point)
				focus_ranges.append(end_point)
		# Add 1.0 at the end
		if focus_ranges[-1] != 1.0:
			focus_ranges.append(1.0)
		focus_ranges = np.array(focus_ranges)

		# Generate thresholds
		thresholds = np.array([])
		for start_point, end_point in zip(focus_ranges[:-1], focus_ranges[1:]):
			thresholds = np.hstack((thresholds, np.linspace(start_point, end_point, num = num)))

		roc = np.matrix([CalculateTPTN(y_pred, y_orig, th) for th in thresholds])

	import matplotlib.pyplot as plt
	plt.ion()
	plt.figure(figsize = (16.0, 10.0))
	plt.plot(roc[:,0], roc[:,1], color = 'red', linewidth = 2)
	plt.plot((0,1), (0,1), color = 'black', linestyle = '--')
	plt.title('ROC')
	plt.xlabel('False Positive Rate')
	plt.ylabel('True Positive Rate')

	if save_to_file:
		for ext in file_exts:
			plt.savefig('{name}_roc.{ext}'.format(name = save_to_file, ext = ext), dpi = 100)

	if not show_plot:
		plt.close()

	# Return crossing point of TP and TN (more accurate with automatic_focus ON)
	# cross_point = thresholds[np.argmin(np.abs(roc[:,0] - roc[:,1]))] # Calculate crosspoint of True Positive and True Negative
	# cross_point_F = thresholds[np.argmin(np.abs(roc[:,2] - roc[:,3]))] # Calculate crosspoint of False Positive and False Negative
	# This should give you best threshold for best True performace or best False perforamce
	# return cross_point

def PlotPredictionsFromFile(predictions_dir, limit = 0, just_all = False, skip_plot = False):
	if os.path.isdir(predictions_dir):
		count = 0
		y_orig_all = []
		y_pred_all = []
		ds_beated = 0
		for file in os.listdir(predictions_dir):
			_, ext = os.path.splitext(file)
			if ext == '.pickle':
				data = None
				with open(os.path.join(predictions_dir, file), 'rb') as f:
					data = pickle.load(f)
				if data:
					# Get the data from pickle
					y_orig 	 = data['y_orig']
					y_pred 	 = data['y_pred']
					name   	 = data['name']
					score 	 = 0
					accuracy = 0

					log.info('Processing dataset \'{name}\''.format(name = name))

					if len(y_orig_all):
						y_orig_all = np.vstack((y_orig_all, y_orig))
						y_pred_all = np.vstack((y_pred_all, y_pred))
					else:
						y_orig_all = y_orig
						y_pred_all = y_pred

					if not just_all:
						# Score and Accuracy doesn't have to be saved in pickle file
						if 'score' in data:
							score  	 = data['score']
							accuracy = data['accuracy']


						file_name = os.path.join(predictions_dir, name)

						# Plot data
						if not skip_plot:
							PlotPredictions(y_pred, y_orig, save_to_file = file_name, file_exts = ['png', 'pdf'])
							PlotROC(y_pred, y_orig, num = 200, save_to_file = file_name, file_exts = ['png', 'pdf'])

						if score and accuracy:
							accuracy_beat = 1.0 / len(y_orig) * (len(y_orig) - np.sum(y_orig))
							isBeated = accuracy > accuracy_beat or np.abs(accuracy - accuracy_beat) < 1e-4

							color_beat = Fore.GREEN if isBeated else Fore.RED

							if isBeated:
								ds_beated += 1

							log.info('Score: {}'.format(score))
							log.info(color_beat + 'Accuracy: {}'.format(accuracy) + Style.RESET_ALL)
							log.info('Accuracy to beat: {}'.format(accuracy_beat))

						with open(file_name + '.txt', 'w') as f:
							f.write(json.dumps({
								'time' : str(datetime.strftime(datetime.now(), '%Y%m%d_%H%M%S')),
								'name' : name,
								'score' : score,
								'accuracy' : accuracy,
								'accuracy_to_beat' : accuracy_beat,
								'is_beated' : int(isBeated),
							}))

					count += 1
					if limit != 0 and count >= limit:
						log.info('Exceeded limit for processed files.')
						break

		log.info('-' * 32)
		log.info('Datasets beated: {} / {}'.format(ds_beated, count))
		with open(predictions_dir + '.info.txt', 'w') as f:
			f.write(json.dumps({
				'time' : str(datetime.strftime(datetime.now(), '%Y%m%d_%H%M%S')),
				'datasets_beated' : ds_beated,
				'datasets_total' : count,
			}))
		if not skip_plot:
			PlotROC(y_pred_all, y_orig_all, num = 200, save_to_file = predictions_dir + '.roc', file_exts = ['png', 'pdf'])

################################################################
# TASK's ACTION
################################################################
def PlotBenchmark(predictions_dir, limit = 0, just_all = False, skip_plot = False):
	def ActionPlotBenchmark(task, error):
		PlotPredictionsFromFile(predictions_dir, limit, just_all, skip_plot)
		return NoError()
	return ActionPlotBenchmark

def BenchmarkModel(task, error):
    from rudy.data_h5_storages import HStorage

    for need in ['model', 'model_id']:
        if need not in task.context:
            return ContextIsEmptyError('Expected \'{need}\' in task context, maybe you forgot to call action that creates this context.'.format(need = need))

    storage = HStorage(task.path, use_cache = True, cache_path = task.cache_path)

    GetAllPredictions(task.context['model'], storage, batch_input_shape = task.batch_input_shape, root = task.experiments_path, calculate_stats = True, save_to_file = True, plot_comparison = False, model_id = task.context['model_id'])

    return NoError()
