# Tasks for Training DNN
# Other
import numpy as np
import os
import json

# Tasks
from rudy.utils_tasks import NoError, ContextIsEmptyError, FileNotExistsError

# Rudy
from rudy.dnn_generators import Plotter, MoniMonitor
from rudy.utils_datasets import GetDatasets
import rudy.utils_logging as log

# Keras
from keras.optimizers import Adamax, SGD
from keras.callbacks import Callback, ModelCheckpoint, EarlyStopping

def LoadDatasets(batch_name, train_size = 0.7):
    def SubLoadDatasets(task, error):
        '''Action that prepares datasets and save it to context.'''
        ds = GetDatasets(task.path, batch_names = batch_name, train_size = train_size, cache_path = task.cache_path, batch_input_shape = task.batch_input_shape) # Train size 1.0 == 100%

        log.info('-' * 32)
        log.info('{} train samples'.format(ds.train.size))
        log.info('{} test samples'.format(ds.test.size))

        # Save context
        task.context['dataset'] = ds
        task.context['batch_name'] = ds.batch_name

        return NoError()
    return SubLoadDatasets

def CompileModel(task, error):
    # Check the context for everything you need
    for need in ['model']:
        if need not in task.context:
            return ContextIsEmptyError('Expected \'{need}\' in task context, maybe you forgot to call action that creates this context.'.format(need = need))

    model = task.context['model']

    optimizer = None
    if task.optimizer == 'adamax':
        optimizer = Adamax(lr = task.learn_rate, beta_1 = 0.9, beta_2 = 0.999, epsilon = 1e-08)
    else:
        optimizer = SGD(lr = task.learn_rate, decay = 1e-6, momentum = 0.9, nesterov = True)

    model.compile(
        loss = 'binary_crossentropy',
        optimizer = optimizer,
        metrics = ['accuracy'])

    task.context['model'] = model
    task.context['model_compiled'] = True
    return NoError()

def TrainModelCustom(task, error):
    '''Train model using train_on_batch'''
    for need in ['dataset', 'model', 'model_id', 'batch_name']:
        if need not in task.context:
            return ContextIsEmptyError('Expected \'{need}\' in task context, maybe you forgot to call action that creates this context.'.format(need = need))

    ds = task.context['dataset']
    model = task.context['model']

    ds_train = ds.train
    ds_test = ds.test

    # Prepare save_file name
    save_file = FindFilename(root = task.experiments_path, subfolder = 'Weights', name = '{}.{}.h5'.format(task.weights_file, task.context['model_id']), timefolder = task.init_time)
    CreateFilepathIfNecessary(save_file)
    SaveCommentFile(save_file, json.dumps({
        'path'       : task.path,
        'batch_name' : task.context['batch_name'],
        'time'       : task.init_time,
        }))

    # Create callbacks
    callbacks = []

    if task.plotter:
        callbacks.append(Plotter(save_to_file = '{}_{}.png'.format(task.path, task.context['batch_name'])))

    if task.moni_server:
        callbacks.append(MoniMonitor(dataset = ds, root = task.moni_server))

    if task.train_checkpoint:
        name, ext = os.path.splitext(save_file)
        callbacks.append(ModelCheckpoint(name + '.{epoch:02d}-{val_acc:.2f}' + ext, monitor = 'val_acc', verbose = 0, save_weights_only = True))

    if task.train_early_stop:
        callbacks.append(EarlyStopping(monitor = 'val_acc', patience = 4, verbose = 0))

    # ----------------------------
    # TRAINING:
    # ----------------------------

    train_size = ds_train.size
    for i in range(task.epochs):
        # Starting next Epoch
        log.info('Epoch {}/{}'.format(i, task.epochs))

        for (X, y) in ds_train.generator:
            loss, acc = model.train_on_batch(X, y)




    # ----------------------------
    # TESTING:
    # ----------------------------

def TrainModel(task, error):
    # Check the context for everything you need
    for need in ['dataset', 'model', 'model_id', 'batch_name', 'model_compiled']:
        if need not in task.context:
            return ContextIsEmptyError('Expected \'{need}\' in task context, maybe you forgot to call action that creates this context.'.format(need = need))

    ds = task.context['dataset']
    model = task.context['model']

    ds_train = ds.train
    ds_test =  ds.test

    class ResetStatesCallback(Callback):
        def __init__(self):
            self.counter = 0
            self.reset_counter = 1
            self.maxlen = GetLSTMResetLen(task, ds_train, 0)
        def reset(self):
            self.maxlen = GetLSTMResetLen(task, ds_train, self.reset_counter)
            # print('Reseting state after: {}, maxlen: {}, reset count: {}'.format(self.counter, self.maxlen, self.reset_counter))
            self.counter = 0
            self.model.reset_states()
            self.reset_counter += 1
        def on_epoch_end(self, epoch, logs={}):
            self.reset_counter = 0
            self.reset()
        def on_batch_end(self, batch, logs={}):
            if self.maxlen:
                if self.counter == self.maxlen: # == is faster then % we can use == because we're reseting self.counter
                    self.reset()
                self.counter += 1

    # Prepare save_file name
    save_file = FindFilename(root = task.experiments_path, subfolder = 'Weights', name = '{}.{}.h5'.format(task.weights_file, task.context['model_id']), timefolder = task.init_time)
    CreateFilepathIfNecessary(save_file)
    SaveCommentFile(save_file, json.dumps({
        'path'       : task.path,
        'batch_name' : task.context['batch_name'],
        'time'       : task.init_time,
        'model_id'   : task.context['model_id'],
        'datasets'   : ds.names,
        'sizes'      : [int(ds.sizes[0]), int(ds.sizes[1])],
        'num_positives' : int(ds.num_positives),
        'num_negatives' : int(ds.num_negatives),
        'config'     : task.config,
        }))

    # Create callbacks
    callbacks = [ResetStatesCallback()] if not task.context['model_id'] == 'Simple_Forward' else []

    if task.plotter:
        callbacks.append(Plotter(save_to_file = '{}_{}.png'.format(task.path, task.context['batch_name'])))

    if task.moni_server:
        callbacks.append(MoniMonitor(dataset = ds, root = task.moni_server))

    if task.train_checkpoint:
        name, ext = os.path.splitext(save_file)
        callbacks.append(ModelCheckpoint(name + '.{epoch:02d}-{acc:.2f}' + ext, monitor = 'acc', verbose = 0, save_weights_only = True))

    if task.train_early_stop:
        callbacks.append(EarlyStopping(monitor = 'acc', patience = task.train_early_stop, verbose = 0))

    if task.fit_using_generator:
        model.fit_generator(
            ds_train.generator,
            nb_epoch          = task.epochs,
            samples_per_epoch = ds_train.size,
            validation_data   = ds_test.generator,
            nb_val_samples    = ds_test.size,
            class_weight      = task.class_weight,
            verbose           = 1,
            callbacks         = callbacks)
    else:
        X_train = ds_train.frames_all
        y_train = ds_train.labels_all

        X_test = ds_test.frames_all
        y_test = ds_test.labels_all

        # Normalize

        model.fit(
            X_train, y_train,
            nb_epoch        = task.epochs,
            batch_size      = ds_train.batch_size,
            validation_data = (X_test, y_test),
            class_weight    = task.class_weight,
            shuffle         = task.shuffle,
            verbose         = 1,
            callbacks       = callbacks)

    model.save(save_file, overwrite = False)

    task.context['model'] = model

    return NoError()

def LoadTrainedModel(desired_model_id = ''):
    def ActionLoadTrainedModel(task, error):
        for need in ['model', 'model_compiled']:
            if need not in task.context:
                return ContextIsEmptyError('Expected \'{need}\' in task context, maybe you forgot to call action that creates this context.'.format(need = need))

        model = task.context['model']

        model_id = desired_model_id or task.context['model_id']

        # Get last file created from same storage and batch
        save_file = FindFilename(root = task.experiments_path, subfolder = 'Weights', name = '{}.{}.h5'.format(task.weights_file, model_id), find_last = True, use_timefolder = True)

        print(save_file)

        if os.path.isfile(save_file):
            model.load_weights(save_file, by_name = task.load_by_name)
            task.context['model'] = model
        else:
            return FileNotExistsError('Weights cannot be loaded, beacause file \'{}\' doesn\'t exists.'.format(save_file), isCritical = False)

        return NoError()
    return ActionLoadTrainedModel

def EvaluateModel(task, error):
    '''Action that evaluates model on current dataset (by context).

    This action wasn't tested, for benchmarking purposes use Benchmark or ShowBenchmark actions, see 'benchmark_datasets.py'.'''
    for need in ['dataset', 'model']:
        if need not in task.context:
            return ContextIsEmptyError('Expected \'{need}\' in task context, maybe you forgot to call action that creates this context.'.format(need = need))

    ds    = task.context['dataset']
    model = task.context['model']

    score, accuracy = model.evaluate_generator(
        ds.all.generator,
        val_samples = ds.all.size
    )

    # TODO: Print accuracy here?
    log.info('Test score:', score)
    log.info('Test accuracy:', accuracy)

    task.context['score']    = score
    task.context['accuracy'] = accuracy

    return NoError()

def PredictWithModel(task, error):
    '''Action that generate prediction for given dataset and model (by context)

    This action wasn't tested, for benchmarking purposes use Benchmark or ShowBenchmark actions, see 'benchmark_datasets.py'.'''
    for need in ['dataset', 'model', 'batch_name']:
        if need not in task.context:
            return ContextIsEmptyError('Expected \'{need}\' in task context, maybe you forgot to call action that creates this context.'.format(need = need))

    ds = task.context['dataset']
    model = task.context['model']

    y_orig = ds.all.labels
    y_pred = model.predict_generator(
        ds.all.generator,
        val_samples = ds.all.size
    )

    from benchmark_datasets import PlotPredictions

    # Show plot here?
    PlotPredictions(y_pred, y_orig, save_to_file = task.context['batch_name'], file_exts = ['png', 'pdf'])

    task.context['y_pred'] = y_pred

    return NoError()

# Helper
def FindBasepath(root, use_timefolder = False, timefolder = '', subfolder = '', find_last = False):
    '''Function returning availible path for saving images, comments, weights etc...'''
    root = os.path.abspath(root)

    # Find last timefolder
    if find_last and use_timefolder:
        dirs = np.sort(os.listdir(root))
        # Last item in listdir should be last timefolder, because we're using time and date to derive timefolder name, however it's not generic, maybe change to os.path.getmtime(...)
        timefolder = dirs[-1]

    # Use timefolder will be turn on automatically, if timefolder isn't empty
    # use_timefolder = use_timefolder or not not timefolder

    # Prepare base path
    return os.path.join(root, timefolder, subfolder)

def FindFilename(root, name, use_timefolder = False, timefolder = '', subfolder = '', find_last = False):
    '''
    Function returning a file path for given name, use it for saving images, weights, comments, etc...

    If find_last is False, it returns next availible filename in timefolder.
    If find_last is True, it tries to find last filename, if use_timefolder is True, it will search the latest timefolder in the root
    '''
    filename, ext = os.path.splitext(os.path.basename(name))

    base_path = FindBasepath(root, use_timefolder = use_timefolder, timefolder = timefolder, subfolder = subfolder, find_last = find_last)
    base_path = os.path.join(base_path, filename)

    count = 0
    while os.path.exists('{base}_{count}{ext}'.format(base = base_path, count = count, ext = ext)):
        count += 1
    if find_last:
        count -= 1
    return '{base}_{count}{ext}'.format(base = base_path, count = count, ext = ext) if count >= 0 else os.path.dirname(base_path)

def CreateFilepathIfNecessary(filepath):
    '''Create all path in the file path if necessary.'''
    import os

    dirs = os.path.dirname(filepath)
    if not os.path.exists(dirs):
        os.makedirs(dirs)
    return filepath

def GetLSTMResetLen(task, ds = None, reset_counter = 0):
    '''Function returning lenght of sequence in given dataset for given reset counter.
    
    It takes configuration value 'lstm_reset':
        If it's tuple of two numbers, it generates random number in the range specified by tupple
        If it's number, it returns the number itself.
        If it's 'auto' you have to provide all arguments (ds, reset_counter) and the function will return lenght of the sequence in the dataset.
    '''
    reset_len = task.lstm_reset
    if type(reset_len) is tuple:
        return np.random.randint(reset_len[0], reset_len[1])
    elif type(reset_len) is int:
        return reset_len
    elif type(reset_len) is str and reset_len == 'auto':
        if not ds or reset_counter >= len(ds.sizes):
            return 0
        else:
            return int(ds.sizes[reset_counter] / ds.batch_size)
    else:
        assert False, 'lstm_reset has to be specific Int or Tuple (min, max) range for generating random number.'

def SaveCommentFile(filepath, content):
    '''Save comment file.'''
    filepath = os.path.splitext(filepath)[0] + '.txt'
    with open(filepath, 'w') as f:
        f.write(content)