import os
import csv
import numpy as np

from .protocols import Printable
from . import utils_logging as log
from .utils_measure import Measure
from .utils_concurrent import Pool

# Key names (key has to match file names)
key_alerts = 'DataAlerts'
key_cfar   = 'DataCFAR'
key_fft    = 'DataFFTPreNRL'
key_freq   = 'DataFreqList'
key_objs   = 'DataObjectList'
key_tracks = 'DataTrackList'

key_side_left  = 'REAR_LEFT'
key_side_right = 'REAR_RIGHT'

idx_alert_BSD_VRaw  = 5
idx_alert_BSD_VMask = 6
idx_alert_BSD_ORaw  = 7
idx_alert_BSD_OMask = 8

idx_alert_CVW_VRaw  = 1
idx_alert_CVW_VMask = 2
idx_alert_CVW_ORaw  = 3
idx_alert_CVW_OMask = 4

idx_fft_beam  = 1;
idx_fft_chirp = 2;
idx_fft_chan  = 3;
idx_fft_start = 4; # Odd indexes = real part, Even = imaginary
idx_fft_end   = -1;

class RDMatrix (Printable):

    def __init__(self, fft_data = None):

        self._phi = {}
        self._amp = {}
        self.future = None

        if Pool.isEnabled() and fft_data:
            self.future = Pool.submit(self._getSpectrum, fft_data[:])
        elif fft_data:
            self._phi, self._amp = self._getSpectrum(fft_data)

    # TODO: Use Multithread?
    def _getSpectrum(self, fft_data):
        _phi = {}
        _amp = {}
        phi = {}
        amp = {}
        # For every row in fft
        for fft_row in fft_data:
            ib  = int(fft_row[idx_fft_beam]) # Beam index
            ich = int(fft_row[idx_fft_chan]) # Channel index
            part_re = np.array([float(f) for f in fft_row[idx_fft_start + 1::2]]) # Every odd number
            part_im = np.array([float(f) for f in fft_row[idx_fft_start::2]])     # Every even number

            temp_phi = np.arctan2(part_re, part_im)
            temp_amp = np.sqrt(np.add(np.square(part_re), np.square(part_im)))

            if ib not in _phi:
                _phi[ib] = {}
                _amp[ib] = {}

            if ich not in _phi[ib]:
                _phi[ib][ich] = temp_phi
                _amp[ib][ich] = temp_amp
            else:
                _phi[ib][ich] = np.vstack((_phi[ib][ich], temp_phi))
                _amp[ib][ich] = np.vstack((_amp[ib][ich], temp_amp))

        # Convert inner dictionary to numpy array, it would looks the same, but it would be more easy to save it
        for ib in _phi:
            phi[ib] = np.array(list(_phi[ib].values()))
            amp[ib] = np.array(list(_amp[ib].values()))

        return phi, amp

    def _processMatrix(self, val, normalize = False, shift = True): 
        '''Post-process matrix - normalize, shift, etc...'''
        # Create matrix
        mat = val;

        # Height of matrix
        h = len(mat)

        # Shift by half of height
        if shift:
            mat = np.roll(mat, int(h / 2), axis = 0)

        # Normalize
        if normalize:
            mat = np.divide(mat, np.amax(mat)) # Should we normalize here?

        return mat

    @property
    def amp(self):
        if not self._amp and self.future:
            # Wait for process to complete and return the result
            self._phi, self._amp = self.future.result()
        return self._amp
    
    @property
    def phi(self):
        if not self._phi and self.future:
            # Wait for process to complete and return the result
            self._phi, self._amp = self.future.result()
        return self._phi
    

    def get(self, beam = 0, channel = None):
        '''Get matrix for specific Beam and Channel'''

        if channel is None:
            m_amp0 = self._processMatrix(self.amp[beam][0])
            m_phi0 = self._processMatrix(self.phi[beam][0])

            m_amp1 = self._processMatrix(self.amp[beam][1])
            m_phi1 = self._processMatrix(self.phi[beam][1])

            return np.array([m_amp0, m_amp1, m_phi0 - m_phi1])
        else:
            m_amp = self._processMatrix(self.amp[beam][channel])
            m_phi = self._processMatrix(self.phi[beam][channel])

            return np.array([m_amp, m_phi])

    def printme(self, simple = False):
        import matplotlib.pyplot as plt

        self.get(0, 0)

        if simple:
            plt.imshow(self.get(0, 0), interpolation="nearest")
            plt.show()
            return

        for b in range(0, len(self.amp)):
            for ch in range(0, len(self.amp[0])):
                # Plot the image
                mat = self.get(b, ch)
                plt.imshow(mat[0], interpolation="nearest")
                plt.show()
                plt.imshow(mat[1], interpolation="nearest")
                plt.show()

class Frame (Printable):
    '''Object representing one Frame (one MCC)'''

    def __init__(self, root, side, mcc):
        self.side = side
        self.mcc  = mcc

        self._path  = os.path.join(root, side, mcc)
        self._rdmat = None

    def getData(self, key):
        '''Get Data from CSVs'''
        csvPath = os.path.join(self._path, key + '.csv')
        if os.path.exists(csvPath):
            with open(csvPath, 'r') as csvFile:
                csvData = csv.reader(csvFile, delimiter=',', quotechar='|')
                return [r for r in csvData]
        else:
            log.error('No CSV FILE for given KEY exists.')
            log.debug('Full Path: \'%s\'' % (csvPath))

        return None

    def calculateRDMatrix(self):
        self._rdmat = RDMatrix(self.getData(key_fft))

    @property
    def rdmatrix(self):
        '''Get RDMatrix (lazy initialization)'''
        if self._rdmat is None:
            self.calculateRDMatrix()
        return self._rdmat

    @property
    def positive(self):
        return int(self.getData(key_alerts)[0][idx_alert_BSD_OMask]) is 1
    

    # Printable protocol method
    def printme(self, simple = False):
        '''Print content.'''
        log.info('Side: %s MCC: %d Path: %s' % (self.side, self.mcc, self._path))