import os
import numpy as np
import h5py

from .protocols import Printable, Cachable
from .data_storages import Storage
from . import utils_logging as log
from .data_frame import RDMatrix, Frame
from .utils_measure import Measure

def getCacheDirectory(path = ''):
	_defaultCacheDirectory = '__h5__'
	path = os.path.join(path, _defaultCacheDirectory)
	if not os.path.exists(path) or not os.path.isdir(path):
		os.mkdir(path)
	return path

KEY_ALERTS = 'Alerts'
KEY_FFT    = 'FFTPreNRL'

KEY_SIDE_LEFT  = 'REAR_LEFT'
KEY_SIDE_RIGHT = 'REAR_RIGHT'

KEY_ALERT_BSD_VRAW  = 'BSDAlert_Vehicle_Raw'
KEY_ALERT_BSD_VMASK = 'BSDAlert_Vehicle_Masked'
KEY_ALERT_BSD_ORAW  = 'BSDAlert_OLT_Raw'			# Default label
KEY_ALERT_BSD_OMASK = 'BSDAlert_OLT_Masked'

KEY_ALERT_CVW_VRAW  = 'CVWAlert_Vehicle_Raw'
KEY_ALERT_CVW_VMASK = 'CVWAlert_Vehicle_Masked'
KEY_ALERT_CVW_ORAW  = 'CVWAlert_OLT_Raw'
KEY_ALERT_CVW_OMASK = 'CVWAlert_OLT_Masked'
 
KEY_HRDMATRIX = 'rdmatrix'
ATTR_FRAME_POSITIVE = 'positive'
ATTR_FRAME_PATH = 'path'
ATTR_POSITIVES = 'positives'
ATTR_NEGATIVES = 'negatives'
ATTR_COMPLETED = 'completed'

class HRDMatrix(RDMatrix):
	def __init__(self, root = None, fft_data = None):
		'''ROOT is group in HDF file representing RDMatrix, it should contains groups for each beam (0-3).'''
		super().__init__(fft_data = fft_data)
		self.root = root

		# If root is not None we can load the data from HDF
		if root:
			for ib in root:
				self.amp[int(ib)] = root[ib]['amp']
				self.phi[int(ib)] = root[ib]['phi']

	# def get(self, beam = 0, channel = None):
	# 	if not channel:
	# 		return self.root[str(beam)]
	# 	elif channel in range(0,2):
	# 		return self.root[str(beam)][channel]

	# 	return None

class HFrame(Frame):
	def __init__(self, root, use_cache = True):
		'''ROOT is group in HDF file representing MCC, it should contains groups "rdmatrix" and attributes like "path" or "positive".'''
		self.hgroup = root
		self.use_cache = use_cache
		self.cached = False

		super().__init__(
			root = root.name,						# Path name to root
			side = root.parent.name.split('/')[-1],	# Side label
			mcc  = root.name.split('/')[-1])		# MCC

	def getData(self, key):
		return self.hgroup[key] if key in self.hgroup else []

	@Measure.this('get_rdmatrix')
	def calculateRDMatrix(self):
		if KEY_HRDMATRIX in self.hgroup and ATTR_COMPLETED in self.hgroup[KEY_HRDMATRIX].attrs:
			self._rdmat = HRDMatrix(root = self.hgroup[KEY_HRDMATRIX])
			self.cached = True
			# log.debug('Frame({}): RDMatrix is loaded from cache file.'.format(self.mcc))
		else:
			# We need to calculate the matrix from FFT
			self._rdmat = HRDMatrix(fft_data = self.getData(KEY_FFT))
			self.cached = False
			# log.debug('Frame({}): RDMatrix was calculated from FFT data and saved for future reused.'.format(self.mcc))
	
	@property
	def positive(self):
		if KEY_ALERTS not in self.hgroup and ATTR_FRAME_POSITIVE in self.hgroup.attrs:
			# Old version
			return self.hgroup.attrs[ATTR_FRAME_POSITIVE]
		else:
			return self.hgroup[KEY_ALERTS].attrs[KEY_ALERT_BSD_ORAW] == 1

	def saveCache(self):
		if not self.use_cache or self.cached:
			return False

		# Save RD Matrices
		rdmatrix = self._rdmat
		groupRDM = self.hgroup.require_group(KEY_HRDMATRIX)
		if not ATTR_COMPLETED in groupRDM.attrs:
			for b in range(0,4):
				groupBeam = groupRDM.require_group(str(b))
				amp = rdmatrix.amp[b]
				phi = rdmatrix.phi[b]
				dsetAmp = groupBeam.require_dataset('amp', shape = amp.shape, dtype = amp.dtype)
				dsetPhi = groupBeam.require_dataset('phi', shape = phi.shape, dtype = phi.dtype)
				dsetAmp[:] = amp
				dsetPhi[:] = phi

			# Set group RDM as completed
			groupRDM.attrs[ATTR_COMPLETED] = True

		return True

class HBatchStorage():
	'''Object representing batch in HDF5 file.'''

	def __init__(self, root):
		self.root = root
		self.name = root.name.split('/')[-1]

	def framesForType(self, typePositives):
		'''Return the dict of generators for all sides for either positives or negatatives data.'''
		key = ATTR_POSITIVES if typePositives else ATTR_NEGATIVES

		# TODO: ATTR_POSITIVES or ATTR_NEGATIVES don't have to exis
		resultDict = {}
		# Go through all sides
		for side in self.root.values():
			# Get the list of positives or negatives
			data = side.attrs[key]
			# Create generator of HFrame for specific MCC and side
			resultDict[side.name.split('/')[-1]] = [HFrame(side[mcc]) for mcc in data]

		return resultDict

	def frameWithSideAndMCC(self, side, mcc):
		'''Return specific frame with given side and given mcc'''
		return HFrame(self.root[side][str(mcc)])

	def framesForSide(self, side):
		'''Return all Frames for specific side (as a generator)'''
		return (HFrame(mcc) for mcc in self.root[side].values())

	@property
	def positive(self):
		'''Return the dict of generators for all sides for positives data.'''
		return self.framesForType(True)

	
	@property
	def negative(self):
		'''Return the dict of generators for all sides for negatatives data.'''
		return self.framesForType(False)

	@property
	def side_frames(self):
		'''Return generator for all frames returning tupple (side, frame)'''
		return ((side.name.split('/')[-1], (HFrame(mcc) for mcc in side.values())) for side in self.root.values())

	def clearContent(self):
		pass

class HStorage(Storage):
	'''Class holding the HDF file.'''

	def __init__(self, path, use_cache = False, cache_path = ''):
		self._file = None
		self._cache_path = cache_path
		super(HStorage, self).__init__(path, use_cache = False, cache_path = cache_path)

		# Overwrite the super
		self._use_cache = use_cache

	# Printable protocol method
	def printme(self, simple=False):
		pass

	@property
	def file(self):
		if not self._file:
			# Open HDF5 File
			self._file = h5py.File(os.path.join(getCacheDirectory(self._cache_path), '%s.h5' % self.name), 'a')
		return self._file

	def __len__(self):
		return len(self._file)

	@property
	def batches(self):
		# If cache is allowed or isn't possible to open HDF file
		if self._cache or not self._file.keys():
			return super(HStorage, self).batches

		return (HBatchStorage(batch) for batch in self._file.values())
	
	# Cachable protocol methods
	# TODO: Split this method to keep the responsibility of exporting format on concrete class
	# TODO: Save whole AMP and PHI for RD Matrix not post-process matrices for different beams! We have to convert the old caches again.
	# TODO: Move manipulating with HDF File outside of this class and add version of file to distinguish between old and new format
	def saveCache(self):
		if self._cache and not self._loadedFromCache:
			super(HStorage, self).saveCache()

		if not self._use_cache:
			return False

		for batch in self.batches:
			# Don't save batch, if there is already saved one
			groupBatch = self.file.require_group(batch.name)

			# TODO: New version doesn't have ATTR_COMPLETED flag
			if not ATTR_COMPLETED in groupBatch.attrs:
				log.debug('Saving batch \'{bname}\' to cache...'.format(bname = batch.name))
				
				for (side, frames) in batch.side_frames:
					# positives = []
					# negatives = []

					# Required group for side
					groupSide = groupBatch.require_group(side)

					if not ATTR_COMPLETED in groupSide.attrs:
						for frame in frames:
							# QUICK FIX ##
							# frame._path = os.path.join(batch._path, frame.side, frame.mcc)
							## --------- #

							groupMCC = groupSide.require_group(frame.mcc)

							# Save Path to MCC
							groupMCC.attrs[ATTR_FRAME_PATH]     = frame._path
							groupMCC.attrs[ATTR_FRAME_POSITIVE] = frame.positive

							# Save RD Matrices
							rdmatrix = frame.rdmatrix
							groupRDM = groupMCC.require_group(KEY_HRDMATRIX)
							if not ATTR_COMPLETED in groupRDM.attrs:
								for b in range(0,4):
									data = rdmatrix.get(b)
									groupBeam = groupRDM.require_group(str(b))
									groupBeam['amp'] = rdmatrix.amp[b]
									groupBeam['phi'] = rdmatrix.phi[b]
								# 	dsetBeam = groupRDM.require_dataset(str(b), shape = data.shape, dtype = data.dtype)
								# 	dsetBeam[:] = data

								# Set group RDM as completed
								groupRDM.attrs[ATTR_COMPLETED] = True

							# Save frame to positives or negatives list
							# if frame.positive:
							# 	positives.append(frame.mcc)
							# else:
							# 	negatives.append(frame.mcc)

						# Save positives and negatives MCCs to side group attributes for faster filtering
						# groupSide.attrs[ATTR_POSITIVES] = np.array(positives, dtype = 'S10')
						# groupSide.attrs[ATTR_NEGATIVES] = np.array(negatives, dtype = 'S10')

						# Set side as completed
						groupSide.attrs[ATTR_COMPLETED] = True

				# Set batch as completed
				groupBatch.attrs[ATTR_COMPLETED] = True

			# Clear batch's content, because we don't need it anymore - save memory
			self.file.flush()
			batch.clearContent()

		return True

	def loadCache(self):
		if self._cache:
			return super(HStorage, self).loadCache()

		# print('Path: ' + str(list(self.file.keys())))
			
		# Load batches Paths
		self._batchesPaths = list(self.file.keys())

		return len(self._batchesPaths) > 0