import os
import numpy as np
import itertools

from random import shuffle
from .protocols import Cachable
from .h5_caching import HCache
from . import utils_logging as log
import itertools
from sklearn.utils import shuffle as skshuffle
from .utils_measure import Measure
from tqdm import tqdm
from collections import defaultdict

def _checkConsistency(frames, labels, mcc):
    len_frames = sum([len(fr) for (_,fr) in frames.items()]) / len(frames)
    if len_frames != len(labels) or (any(mcc) and len(labels) != len(mcc)):
        log.error('Frames({len_fr}) and Labels({len_la}){mcc} have to have same dimensions.'.format(
            len_fr = len(frames),
            len_la = len(labels),
            mcc = 'and MCC({})'.format(len(mcc)) if mcc else ''))
        return False
    return True

# Simple data set
class DataSet():

    def __init__(self, frames, labels, mcc = [], num_positives = -1, next_ds = None, batch_input_shape = (1, 16, 64, 3)):
        #TODO: Hotfix
        assert _checkConsistency({'0':frames}, labels, mcc)

        self._frames = frames
        self._labels = labels
        self._mcc = mcc if any(mcc) else np.arange(len(labels))
        self._num_positives = num_positives
        self._next_ds = next_ds
        self._size = 0
        self._sizes = []
        self.batch_input_shape = batch_input_shape

        if batch_input_shape[0] == 1:
            log.warning('Batch size is set to 1, are you sure about it?')

    @classmethod
    def fromPosNeg(cls, positive, negative, mccs = ([], []), doShuffle = False):
        return cls(
            frames = np.vstack((positive,negative)),
            labels = ([1] * len(positive)) + ([0] * len(negative)),
            mcc = np.vstack(mccs),
            num_positives = len(positive))

    def doShuffle(self):
        self.frames, self.labels, self.mcc = skshuffle(self.frames, self.labels, self.mcc, random_state = 0)


    def _expand_transpose(self, x):
        result = x.transpose(0, 2, 3, 1)
        if len(self.batch_input_shape) == len(self._frames.shape) + 1:
            result = np.expand_dims(result, axis = 1)
        elif len(self.batch_input_shape) > len(self._frames.shape):
            assert False, 'Unexpected batch_input_shape: {}'.format(self.batch_input_shape)
        return result

    # ------------------------------------------------------------------
    # Non generative accessors
    # ------------------------------------------------------------------
    @property
    def batch_size(self):
        return self.batch_input_shape[0]

    @property
    def frames(self):
        return self._expand_transpose(self._frames[:len(self)])

    @property
    def frames_all(self):
        return np.vstack([ds.frames for ds in self.generator_datasets(finite = True)])

    @property
    def labels(self):
        return np.matrix(self._labels[:len(self)]).T
        # return np.reshape(np.array(self._labels), (len(self._labels), 1))

    @property
    def labels_all(self):
        return np.vstack([ds.labels for ds in self.generator_datasets(finite = True)])

    @property
    def mcc(self):
        return np.matrix(self._mcc[:len(self)]).T

    @property
    def mcc_all(self):
        return np.vstack([ds.mcc for ds in self.generator_datasets(finite = True)])

    @property
    def next_ds(self):
        '''Return next connected dataset or SELF if there is non or None if 'return_self' is False.'''
        return self._next_ds if self._next_ds else self
    # ------------------------------------------------------------------

    # ------------------------------------------------------------------
    # Generative accessors
    # ------------------------------------------------------------------

    def generator_datasets(self, finite = False):
        '''Returns generator over connected datasets, if finite is True, generator terminates, otherwise it will runs indefinitelly.'''
        ds = self
        while True:
            yield ds
            if finite and ds is ds.next_ds:
                break
            ds = ds.next_ds

    @property
    def generator_finite(self):
        '''Returns FINITE generator over frames and labels, generating tupple (frame, label)
        '''
        # return ((self._expand_transpose(np.expand_dims(frame, axis = 0)), np.matrix(label)) for frame, label in zip(self._frames, self._labels))

        for i in range(int(len(self) / self.batch_size) - self.batch_size):
            frames = self._expand_transpose(self._frames[i * self.batch_size : (i + 1) * self.batch_size])
            labels = np.matrix(self._labels[i * self.batch_size : (i + 1) * self.batch_size]).T

            assert frames.shape[0] == self.batch_size, '{}: Wrong number of frames in batch: {}'.format(i, frames.shape)

            yield (frames, labels)

    @property
    def generator(self):
        '''Function that generates INFINITE generator, for methods fit_generator, predict_generator or evaluate_generator in Keras.'''

        # Get infinite generator over connected datasets
        dss = self.generator_datasets(finite = False)
        # Get finite generator over (frames, labels)
        gen = next(dss).generator_finite # next(dss) returns SELF as a first dataset
        while True:
            try:
                next_gen = next(gen)
                yield next_gen
            except StopIteration:
                # Load generator from next availible dataset
                gen = next(dss).generator_finite

    # ------------------------------------------------------------------

    @property
    def num_positives(self):
        if self._num_positives is -1:
            self._num_positives = np.sum(self._labels[:len(self)])
        return self._num_positives

    @property
    def num_negatives(self):
        return len(self) - self.num_positives

    def __len__(self):
        size = len(self._labels)
        return size - (size % self.batch_size)

    @property
    def sizes(self):
        if not self._sizes:
            self._sizes = [len(ds) for ds in self.generator_datasets(finite = True)]
        return self._sizes

    @property
    def size(self):
        '''Returning size of all connected datasets'''
        if not self._size:
            self._size = np.sum(self.sizes)
        return self._size


class DataSets(Cachable):

    _cache_prefix = ''

    def __init__(self, path = '', use_cache = True, cache_path = '', batch_name = '', choosen_side = 'REAR_LEFT', fixed_beams = [0,1,2,3], train_size = 0.6, batch_input_shape = (1, 16, 64, 3)):
        '''Define DataSets. DataSets class is used for generating Test and Train datasets from given frames.'''
        self.path = path
        self.name = os.path.basename(os.path.normpath(path))
        self.batch_name = batch_name
        self.use_cache = use_cache

        self.batch_input_shape = batch_input_shape
        self.choosen_side = choosen_side
        self.fixed_beams = fixed_beams
        self._train_size = train_size

        self.frames = {}
        self.labels = []
        self.mcc = []

        self._num_positives = 0
        self._num_negatives = 0

        self.cached = False

        if use_cache:
            # Define cache
            self._cache = HCache(
                prefix = '{prefix}{name}_'.format(prefix = self._cache_prefix, name = self.name),
                name = '_{postfix}'.format(postfix = batch_name),
                folder = 'ds',
                path = cache_path)

            if not self.loadCache():
                log.debug('DataSets({side},{beam}) could not be loaded from cache.'.format(side = self.choosen_side, beam = fixed_beams))
                self.cached = False
            else:
                self.cached = True
        else:
            self._cache = None

    def __len__(self):
        '''Return total size.'''
        return len(self.labels)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._cache.close()

    def setData(self, frames, labels = [], mcc = [], analyze = True, overwrite = False):
        '''Convinient method for setting the data either manually of automatically from list of HFrames.

        If you pass just the list of HFrames, it will use the method self.setDataFromFrames(frames).
        If you pass frames and labels, it will use the method self.setDataManually(frames, labels, ...)
        '''
        if not labels and not mcc:
            return self.setDataFromFrames(frames, overwrite)
        elif frames and labels:
            return self.setDataManually(frames, labels, mcc, analyze, overwrite)

    def setDataManually(self, frames, labels, mcc = [], analyze = True, overwrite = False):
        '''Set data frames, labels and possibly mcc.
        Args:
            frames - dictionary of frames, frames per beams
            labels - list of labels corresponding with frames
            mcc    - list of mcc corresponding with frames (optional)
        '''
        assert _checkConsistency(frames, labels, mcc)

        if self.cached:
            log.warning('DataSets({side},{beam}) is already loaded from cache, the data will be overwritten.'.format(side = self.choosen_side, beam = self.fixed_beams))
            if not overwrite:
                return False

        self._num_positives = 0
        self._num_negatives = 0
        self._positives_start_indices = []
        self._positives_ranges = []

        positives_ranges = 0
        positives_start_index = 0
        previous_positive = False

        self.frames = frames
        self.labels = labels if labels.dtype == int else labels.astype(int)
        self.mcc = mcc if mcc else np.empty(len(labels))

        if analyze:
            for i, label in enumerate(self.labels):
                if label:
                    if not previous_positive:
                        positives_start_index = i

                    self._num_positives += 1
                    previous_positive = True
                else:
                    if previous_positive:
                        positives_ranges = i - positives_start_index
                        self._positives_start_indices.append(positives_start_index)
                        self._positives_ranges.append(positives_ranges)

        return True

    def setDataFromFrames(self, frames, overwrite = False):
        '''Set data using frames.
        Args:
            frames - list of frames (where Frame is instance of Frame class)
        '''
        if self.cached:
            log.warning('DataSets({side}) is already loaded from cache, the data will be overwritten.'.format(side = self.choosen_side))
            if not overwrite:
                return False

        tmp_frames = defaultdict(list)
        tmp_rdmatrices = []
        tmp_labels = []
        tmp_mcc = []

        self._num_positives = 0
        self._num_negatives = 0
        self._positives_start_indices = []
        self._positives_ranges = []

        positives_ranges = 0
        positives_start_index = 0
        previous_positive = False

        for i, frame in enumerate(tqdm(frames)):
            Measure.start('saving_frame')
            tmp_rdmatrices.append(frame.rdmatrix)
            tmp_labels.append(int(frame.positive))
            tmp_mcc.append(int(frame.mcc))

            if frame.positive:
                if not previous_positive:
                    positives_start_index = i

                self._num_positives += 1
                previous_positive = True
            else:
                if previous_positive:
                    positives_ranges = i - positives_start_index
                    self._positives_start_indices.append(positives_start_index)
                    self._positives_ranges.append(positives_ranges)

                self._num_negatives += 1
                previous_positive = False

            Measure.end('saving_frame')

        # To take advantage of concurrent calculation of RDMatrix we need to first get the rdmatrix which invokes the calculation so we can get the beam matrix later
        for rdmatrix in tqdm(tmp_rdmatrices):
            for beam in self.fixed_beams:
                tmp_frames[str(beam)].append(np.array(rdmatrix.get(beam)))

        self.frames = tmp_frames
        self.labels = np.array(tmp_labels)
        self.mcc = np.array(tmp_mcc)

        return True

    @property
    def train_size(self):
        train_size = self._train_size
        if train_size <= 1.0:
            train_size = len(self) * train_size
        return int(train_size)

    @property
    def names(self):
        return [self.name]

    @property
    def sizes(self):
        '''Return tupple with train and test sizes (train_size, test_size).'''
        train_size = self.train_size
        return (train_size, len(self) - train_size)

    @property
    def all(self):
        '''Return DataSet with all data'''
        #TODO: Figure out how to get specified beam
        return DataSet(
            frames = self.frames[str(self.fixed_beams[0])],
            labels = self.labels,
            mcc = self.mcc,
            batch_input_shape = self.batch_input_shape)

    @property
    def train(self):
        '''Return DataSet with Train data'''
        #TODO: Figure out how to get specified beam
        return DataSet(
            frames = self.frames[str(self.fixed_beams[0])][:self.train_size],
            labels = self.labels[:self.train_size],
            mcc = self.mcc[:self.train_size],
            batch_input_shape = self.batch_input_shape)

    @property
    def test(self):
        '''Return DataSet with Test data'''
        #TODO: Figure out how to get specified beam
        return DataSet(
            frames = self.frames[str(self.fixed_beams[0])][self.train_size:],
            labels = self.labels[self.train_size:],
            mcc = self.mcc[self.train_size:],
            batch_input_shape = self.batch_input_shape)

    @property
    def num_positives(self):
        '''Return number of positives data in the whole datasets'''
        num = self._num_positives
        if not num:
            num = np.sum(self.labels)
        return num

    @property
    def num_negatives(self):
        '''Return number of negatives data in the whole datasets'''
        num = self._num_negatives
        if not num:
            num = len(self.labels) - np.sum(self.labels)
        return num

    def loadCache(self):
        '''Load DataSets from given path, if given path doesn't exists, return False otherwise return True.'''
        if self._cache is None:
            return False

        d = self._cache.load(group = '/{side}'.format(side = self.choosen_side))
        if d is not None:
            self.setData(
                frames = d['frames'],
                labels = d['labels'],
                mcc = d['mcc'],
                analyze = False) # Analyze = False will disable post-processing of data, we have all post-processed data in Cache.attrs

            self._num_positives = d.attrs['num_positives']
            self._num_negatives = d.attrs['num_negatives']
            self._positives_start_indices = d.attrs['positives_start_indices']
            self._positives_ranges = d.attrs['positives_ranges']

            log.info('Total number of frames loaded from cache: {}'.format(len(self.frames)))
        else:
            log.warning('Cannot load DataSets from cache.')
            return False
        return True

    def saveCache(self):
        if self._cache is None:
            return False

        log.info('Saving DataSets to cache...')
        if not self._cache.save({
            'frames' : self.frames,
            'labels' : self.labels,
            'mcc' : self.mcc
        }, attrs = {
            'name' : self.name,
            'path' : self.path,
            'choosen_side' : self.choosen_side,
            'fixed_beams' : self.fixed_beams,
            'num_positives' : self.num_positives,
            'num_negatives' : self.num_negatives,
            'positives_start_indices' : self._positives_start_indices,
            'positives_ranges' : self._positives_ranges,
        }, group = '/{side}'.format(side = self.choosen_side), overwrite = True):
            log.warning('Cannot save DataSets to cache.')
            return False
        else:
            log.info('Succesfully saved DataSets to cache.')
        return True

# Collection of data sets
class SplitDataSets(DataSets):

    _cache_prefix = 's_'

    def __init__(self, path = '', use_cache = True, choosen_side = 'REAR_LEFT', fixed_beams = 2, train_size = 0.6, pos_neg_ratio = 1, positive = [], negative = [], cache_path = '', positive_skip = 1, negative_skip = 1):

        self.path = path
        self.name = os.path.basename(os.path.normpath(path))
        self.choosen_side  = choosen_side
        self.fixed_beams    = fixed_beams
        self.train_size    = train_size
        self.pos_neg_ratio = pos_neg_ratio

        self._positive = []
        self._negative = []
        self.pos_train_size = 0
        self.neg_train_size = 0

        self.positive_skip = positive_skip
        self.negative_skip = negative_skip

        if positive or negative:
            self.setData(positive, negative)

        if use_cache:
            self._cache = HCache(prefix = '%s%s_' % (self._cache_prefix, self.name), name = '%s_B%s' % (choosen_side, str(fixed_beams)), folder = 'ds', path = cache_path)
        else:
            self._cache = None

    def __len__(self):
        return len(self.positive) + len(self.negative)

    def setData(self, positive, negative):
        self._positive = positive
        self._negative = negative
        self.pos_train_size, self.neg_train_size = self.calculateSizes(self.train_size, self.pos_neg_ratio)

        log.info('Training dataset size: %d / %d (positive / negative)' % (self.pos_train_size, self.neg_train_size))

    def calculateSizes(self, train_size, pos_neg_ratio):
        if train_size < 1:
            train_size = int(len(self._positive) * train_size)
        return int(min(train_size, len(self._positive))), int(min(train_size / pos_neg_ratio, len(self._negative)))

    def doShuffle(self):
        shuffle(self._positive)
        shuffle(self._negative)

    @property
    def size(self):
        total = self.num_positives + self.num_negatives
        train_size = self.pos_train_size + self.neg_train_size
        return train_size, total - train_size

    @property
    def train(self):
        return DataSet.fromPosNeg(
            positive = self._positive[:self.pos_train_size],
            negative = self._negative[:self.neg_train_size])

    @property
    def test(self):
        return DataSet.fromPosNeg(
            positive = self._positive[self.pos_train_size::self.positive_skip],
            negative = self._negative[self.neg_train_size::self.negative_skip])

    @property
    def num_positives(self):
        return len(self._positive)

    @property
    def num_negatives(self):
        return len(self._negative)

    @property
    def positive(self):
        return self._positive

    @property
    def negative(self):
        return self._negative

    def loadCache(self):
        '''Load DataSets from given path, if given path doesn't exists, return False otherwise return True.'''
        if self._cache is None:
            return False

        # print(self._cache.fullPath)

        d = self._cache.load()
        if d is not None:
            self.setData(
                positive = d['positive'],
                negative = d['negative'])

            log.info('Total number of data loaded from cache: %d / %d (positive / negative)' % (len(self._positive), len(self._negative)))
        else:
            log.warning('Cannot load DataSets from cache.')
            return False
        return True

    def saveCache(self):
        if self._cache is None:
            return False

        # print('NAME: ' + self._cache.name)
        # print('PATH: ' + self._cache.fullPath)

        log.info('Saving DataSets to cache...')
        if not self._cache.save({
            'positive' : self._positive,
            'negative' : self._negative
        }):
            log.warning('Cannot save DataSets to cache.')
            return False
        else:
            log.info('Succesfully saved DataSets to cache.')
        return True

# Multiple datasets
class MultiDataSets(DataSets):

    def __init__(self, datasets = [], name = ''):
        self._datasets = datasets

        self.name = name if name else '(' + '|'.join([ds.name[:6] for ds in self._datasets]) + ')'
        self.batch_name = '[' + '|'.join([ds.batch_name[:6] for ds in self._datasets]) + ']'

        self._dss_all = None
        self._dss_train = None
        self._dss_test = None

        self.choosen_side = datasets[0].choosen_side
        self.fixed_beams  = datasets[0].fixed_beams

        for ds in datasets[1:]:
            assert self.choosen_side == ds.choosen_side and self.fixed_beams == ds.fixed_beams, 'All datasets have to have same choosen_side and fixed_beams.'

    @classmethod
    def fromDataSets(cls, *datasets, name = ''):
        return cls(*datasets, name)

    @property
    def names(self):
        return ['{}.{}'.format(ds.name, ds.batch_name) for ds in self._datasets]

    @property
    def train_size(self):
        return np.sum([ds.train_size for ds in self._datasets])

    @property
    def sizes(self):
        return tuple(np.sum([ds.sizes for ds in self._datasets], axis = 0))

    @property
    def all(self):
        if not self._dss_all:
            # Use Lazy initialization
            dss = [ds.all for ds in self._datasets]
            for i in range(len(dss) - 1):
                dss[i]._next_ds = dss[i + 1]
            self._dss_all = dss[0]
        return self._dss_all

    @property
    def train(self):
        if not self._dss_train:
            # Use Lazy initialization
            dss = [ds.train for ds in self._datasets]
            for i in range(len(dss) - 1):
                dss[i]._next_ds = dss[i + 1]
            self._dss_train = dss[0]
        return self._dss_train

    @property
    def test(self):
        if not self._dss_test:
            # Use Lazy initialization
            dss = [ds.test for ds in self._datasets]
            for i in range(len(dss) - 1):
                dss[i]._next_ds = dss[i + 1]
            self._dss_test = dss[0]
        return self._dss_test

    @property
    def num_positives(self):
        return np.sum([ds.num_positives for ds in self._datasets])

    @property
    def num_negatives(self):
        return np.sum([ds.num_negatives for ds in self._datasets])
