import os

from .data_frame import Frame
from .caching import Cache
from .protocols import Printable, Cachable
from . import utils_logging as log
from tqdm import tqdm

from collections import defaultdict
from itertools import chain

class BatchStorage(Printable, Cachable):
	"""Batch Storage is storage for each batch of CSVs data already converted to Frames and filtered to positives and negatives."""

	_cache_prefix = 'b_'

	def __init__(self, path = None, parent = None, use_cache = True, cache_path = ''):
		
		self._path 	  = os.path.normpath(path)
		self._name 	  = os.path.basename(path)
		self._parent  = parent
		self._positive = defaultdict(list)
		self._negative = defaultdict(list)
		self._frames   = defaultdict(list)

		if use_cache:
			self._cache = Cache(path = cache_path, prefix = self._cache_prefix, name = self._name, folder = parent)
		else:
			self._cache = None

		if not self.canLoadFromCache():
			self.loadFromCSV()
			if self.saveCache():
				# To save memory usage delete positive and negative
				self.clearContent()

	@property
	def positive(self):
		if not self._positive:
			self.loadCache()
		
		positive = defaultdict(list)
		for side in self._positive:
			positive[side] = self._frames[side][self._positive]
		return positive

	@property
	def negative(self):
		if not self._negative:
			self.loadCache()
		
		negative = defaultdict(list)
		for side in self._negative:
			negative[side] = self._frames[side][self._negative]
		return negative

	@property
	def frames(self):
		if not self._frames:
			self.loadCache()
		return self._frames

	def framesForSide(self, side):
		'''Return all Frames for given side (as an array)'''
		return self._frames[side]

	@property
	def side_frames(self):
	    return ((side, p + n) for ((side,p),(_,n)) in zip(self.positive.items(), self.negative.items()))

	@property
	def name(self):
	    return self._name
	
	def clearContent(self):
		self._frames.clear()
		self._positive.clear()
		self._negative.clear()

	def loadFromCSV(self):
		log.info('Processing Batch: "%s"' % os.path.basename(self._path))

		side_gen = (s for s in os.listdir(self._path) if not s.startswith('.'))
		for side in side_gen:

			log.debug('Start processing side: %s' % side)

			mcc_gen = [m for m in os.listdir(os.path.join(self._path, side)) if not m.startswith('.')]

			for mcc in tqdm(mcc_gen):
				# Load frame
				frame = Frame(self._path, side, mcc)

				# Calculate RDMatrix for frame
				frame.calculateRDMatrix()

				# Save frame as positive or negative
				self._frames[side].append(frame)

				if frame.positive:
					self._positive[side].append(len(self._frames[side]) - 1)
				else:
					self._negative[side].append(len(self._frames[side]) - 1)

			log.info('For side "%s", finish processing %d MCCs.' % (side, len(mcc_gen)))
			log.info('Batch contains: %d / %d (positive / negative frames).' % (len(self.positive[side]), len(self.negative[side])))

	# Printable Protocol method
	def printme(self, simple = False):
		log.info('[POSITIVE]:')
		log.info('------------------------')
		for f in self.positive:
			f.printme(simple)
		log.info('')
		log.info('[NEGATIVE]:')
		log.info('------------------------')
		for f in self.negative:
			f.printme(simple)

	# Cachable Protocol methods
	def saveCache(self):
		if self._cache is None:
			return False

		if not self._cache.save({
			'frames'   : self._frames,
			'positive' : self._positive,
			'negative' : self._negative
			}):
			log.warning('Cannot save batch \'%s\' to cache.' % self._name)
			return False
		else:
			log.info('Succesfully saved batch \'%s\' to cache.' % self._name)

		return True

	def canLoadFromCache(self):
		if self._cache is None:
			return False

		if not self._cache.canLoad():
			log.debug('Batch \'%s\' CANOT be loaded from cache.' % self._name)
			return False

		log.debug('Batch \'%s\' CAN be loaded from cache.' % self._name)
		return True


	def loadCache(self):
		if self._cache is None:
			return False

		d = self._cache.load()
		if d is None:
			log.warning('Cannot load batch \'%s\' from cache.' % self._name)
			return False
		else:
			log.info('Succesfully loaded batch \'%s\' from cache.' % self._name)

		self._frames   = d['frames']
		self._positive = d['positive']
		self._negative = d['negative']

		return True

class Storage(Printable, Cachable):

	_cache_prefix = 's_'

	def __init__(self, path, use_cache = True, cache_path = ''):
		self._batchesPaths = []
		self._batches 	   = []
		self._path         = os.path.normpath(path)
		self._name         = os.path.basename(self._path)
		self._use_cache    = use_cache

		self._loadedFromCache = True

		if use_cache:
			self._cache = Cache(path = cache_path, prefix = self._cache_prefix, name = self._name)
		else:
			self._cache = None

		# Load Paths from cache or from file system
		if not self.loadCache():
			self._collectPaths()
			self._loadedFromCache = False

		# If cache is enabled, save to cache
		self.saveCache()

		# Gether batches from given paths
		# self._collectBatches()

	def _collectPaths(self):
		self._batchesPaths = [f for f in os.listdir(self._path) if os.path.isdir(os.path.join(self._path, f)) and not f.startswith('.')]
		log.debug('Batches found: ' + str(self._batchesPaths))

	def _collectBatches(self):
		self._batches = (BatchStorage(os.path.join(self._path, batchPath), use_cache = self._cache is not None, parent = self._name) for batchPath in self.batchesPaths)

	@property
	def batchesPaths(self):
		if not self._batchesPaths:
			self._collectPaths()
		return self._batchesPaths

	@property
	def batches(self):
		if not self._batches:
			self._collectBatches()
		return self._batches

	@property
	def name(self):
	    return self._name

	# Printable Protocol method
	def printme(self, simple = False):
		pass

	# Cachable Protocol methods
	def saveCache(self):
		if self._cache is None:
			return False

		if not self._cache.save({
			'batches' : self._batchesPaths
			}):
			log.warning('Cannot save storage \'%s\' to cache.' % self._name)
			return False
		else:
			log.info('Succesfully saved storage \'%s\' to cache.' % self._name)

		# for batch in self._batches:
		# 	batch.saveCache()

		return True

	def loadCache(self):
		if self._cache is None:
			return False

		d = self._cache.load()
		if d is None:
			log.warning('Cannot load storage \'%s\' from cache.' % self._name)
			return False

		log.info('Succesfully loaded storage \'%s\' from cache.' % self._name)
		self._batchesPaths = d['batches']
		return True
