from keras.utils import np_utils
from keras.callbacks import History, Callback

import time
from .data_sets import DataSets
from . import utils_logging as log

class PlotGenerator(object):
	def __init__(self,
				save_to_file = None,
				show_plot_window = True):
		self.save_to_file = save_to_file
		self.show_plot_window = show_plot_window

		self.linestyles = ['r-', 'b-', 'r:', 'b:']
		self.linestyles_first_epoch = ['rs-', 'b^-', 'r:', 'b:']

		import matplotlib.pyplot as plt
		fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(24, 8))
		plt.ion()

		box1, box2 = ax1.get_position(), ax2.get_position()
		ax1.set_position([box1.x0, box1.y0 + box1.height * 0.1, box1.width, box1.height * 0.9])
		ax2.set_position([box2.x0, box2.y0 + box2.height * 0.1, box2.width, box2.height * 0.9])

		self.fig = fig
		self.ax1 = ax1
		self.ax2 = ax2

	def update(self, epoch, train_loss, train_acc, val_loss, val_acc):
		import matplotlib.pyplot as plt
		self._redraw_plot(epoch, train_loss, train_acc, val_loss, val_acc)

		# show plot window or redraw an existing one
		if self.show_plot_window:
			plt.figure(self.fig.number)
			plt.pause(0.1)

		if self.save_to_file:
			self._save_plot(self.save_to_file)

	def _save_plot(self, filepath):
		self.fig.savefig(filepath)

	def _redraw_plot(self, epoch, train_loss, train_acc, val_loss, val_acc):
		import matplotlib.pyplot as plt
		ax1 = self.ax1
		ax2 = self.ax2

		# List of each epoch (x-axis)
		epochs = list(range(0, epoch+1))

		# Clear loss and accuracy charts
		ax1.clear()
		ax2.clear()

		# Set titles of charts (at the top)
		ax1.set_title('loss')
		ax2.set_title('accuracy')

		linestyles = self.linestyles if epoch > 0 else self.linestyles_first_epoch

		# Plot the lines
		if train_loss:
			ax1.plot(epochs, train_loss, linestyles[0], label='train loss')
		if val_loss:
			ax1.plot(epochs, val_loss, linestyles[1], label='val loss')
		if train_acc:
			ax2.plot(epochs, train_acc, linestyles[0], label='train acc')
		if val_acc:
			ax2.plot(epochs, val_acc, linestyles[1], label='val acc')
		
		# Add legend (below chart)
		ax1.legend(['train loss', 'val loss',
					'train loss regression', 'val loss regression'],
				   bbox_to_anchor=(0.7, -0.08), ncol=2)
		ax2.legend(['train acc', 'val acc',
					'train acc regression', 'val acc regression'],
				   bbox_to_anchor=(0.7, -0.08), ncol=2)
		
		# Labels for x and y axis
		ax1.set_ylabel('loss')
		ax1.set_xlabel('epoch')
		ax2.set_ylabel('accuracy')
		ax2.set_xlabel('epoch')
		
		# Show a grid in both charts
		ax1.grid(True)
		ax2.grid(True)

class Plotter(History):
	def __init__(self,
				save_to_file = None,
				show_plot_window = True):
		super(Plotter, self).__init__()
		self.plotGen = PlotGenerator(save_to_file, show_plot_window)
		self.train_loss = []
		self.train_acc  = []
		self.val_loss   = []
		self.val_acc    = []

	def on_epoch_end(self, epoch, logs={}):
		super(Plotter, self).on_epoch_end(epoch, logs)
		dv = self.params['do_validation']

		self.train_loss.append(logs.get('loss'))
		self.val_loss.append(logs.get('val_loss') if dv else 0)
		self.train_acc.append(logs.get('acc'))
		self.val_acc.append(logs.get('val_acc') if dv else 0)

		self.plotGen.update(epoch, self.train_loss, self.train_acc, self.val_loss, self.val_acc)

class MoniMonitor(Callback):
	'''Callback used to stream events to a Moni Server.
	Requires the `requests` library.
	# Arguments
		root: root url to which the events will be sent (at the end
			of every epoch). Events are sent to
			`root + '/publish/epoch/end/'`. Calls are HTTP POST,
			with a `data` argument which is a JSON-encoded dictionary
			of event data.
	'''
	
	def __init__(self, dataset, root = 'http://localhost:9000'):
		from datetime import datetime
		self.root = root
		self.name = dataset.name
		self.batch_name = dataset.batch_name
		self.dataset = dataset
		self.date = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
		self.start_train_time = 0
		self.start_epoch_time = 0
		self.avg_epoch_time = 0

	def _postData(self, endpoint, data):
		import requests
		import json
		try:
			requests.post(self.root + endpoint, {'data' : json.dumps(data)})
		except:
			log.warning('Could not reach Moni server, check if it\'s running on ' + str(self.root))

	def _baseData(self):
		return {
			'dataset'    : self.name,
			'batch_name' : self.batch_name,
			'date'       : self.date,
		}

	def on_train_begin(self, logs = {}):
		# Send begin event to the Moni server
		self.start_train_time = time.time()
		self.start_epoch_time = time.time()

		data = self._baseData()
		data['choosen_side'] 	= self.dataset.choosen_side
		data['fixed_beam'] 		= self.dataset.fixed_beams
		data['train_size'] 		= int(self.dataset.sizes[0])
		data['test_size']		= int(self.dataset.sizes[1])
		data['pos_size']        = int(self.dataset.num_positives)
		data['neg_size']        = int(self.dataset.num_negatives)

		self._postData('/publish/train/begin/', data)

	def on_train_end(self, logs = {}):
		# Send end event to the Moni server
		data = self._baseData()
		data['finish'] = True
		data['choosen_side'] 	= self.dataset.choosen_side
		data['fixed_beam'] 		= self.dataset.fixed_beams
		data['train_size'] 		= int(self.dataset.sizes[0])
		data['test_size']		= int(self.dataset.sizes[1])
		data['pos_size']        = int(self.dataset.num_positives)
		data['neg_size']        = int(self.dataset.num_negatives)
		data['train_time']		= time.time() - self.start_train_time
		data['epoch_time'] = self.avg_epoch_time  
		data['stats'] = {}
		for k, v in logs.items():
			data['stats'][k] = v

		self._postData('/publish/train/end/', data)

	def on_epoch_end(self, epoch, logs={}):
		epoch_time = time.time() - self.start_epoch_time
		self.avg_epoch_time =  epoch_time if not self.avg_epoch_time else ((self.avg_epoch_time + epoch_time) / 2)
		self.start_epoch_time = time.time()

		data = self._baseData()
		data['epoch_time'] = self.avg_epoch_time  
		data['stats'] = {}
		data['stats']['epoch'] = epoch
		for k, v in logs.items():
			data['stats'][k] = v

		self._postData('/publish/epoch/end/', data)

