import os
import h5py
from .caching import Cache
from . import utils_logging as log

ATTR_COMPLETED = 'completed'
OPT_COMPRESSION_METHOD = 'gzip'
OPT_COMPRESSION_RATE   = 4

class HCache(Cache):
	'''HDF File cache'''

	_suffix = '.h5'
	_defaultCacheDirectory = '__h5__'

	def __init__(self, path = '', prefix = '', name = '', folder = None, compress = False):
		''' '''
		super(HCache, self).__init__(path = path, prefix = prefix, name = name, folder = folder)

		self.compress = compress
		self.root = None

	def close():
		if self.root:
			self.root.close()

	def save(self, data, attrs = {}, group = '/', overwrite = False):
		if self.root:
			self.root.close()

		open_mode = 'a'
		if overwrite:
			# Reopen the file and truncate the content
			open_mode = 'w'

		# Save to Cache in isolation mode, to be sure the class is correctly closed
		with h5py.File(self.fullPath, open_mode) as f:

			if not overwrite and ATTR_COMPLETED in f.attrs:
				return False

			g = f.require_group(group)

			# Save datasets
			for (key, value) in data.items():
				log.debug('CACHE: Saving {key}...'.format(key = key))

				if isinstance(value, dict):
					k = g.require_group(key)
					for subkey, subvalue in value.items():
						if self.compress:
							k.create_dataset(subkey, data = subvalue, compression = OPT_COMPRESSION_METHOD, compression_opts = OPT_COMPRESSION_RATE)
						else:
							k.create_dataset(subkey, data = subvalue)
				else:
					if self.compress:
						g.create_dataset(key, data = value, compression = OPT_COMPRESSION_METHOD, compression_opts = OPT_COMPRESSION_RATE)
					else:
						g.create_dataset(key, data = value)

			# Save attributes
			for (key, value) in attrs.items():
				g.attrs[key] = value

			g.attrs[ATTR_COMPLETED] = True

		return True

	def load(self, group = '/'):
		if self.root:
			self.root.close()

		if not os.path.isfile(self.fullPath):
			return None

		self.root = h5py.File(self.fullPath, 'r')

		if group not in self.root:
			return None

		g = self.root[group]

		if not g.keys() or ATTR_COMPLETED not in g.attrs:
			return None

		return g




