import os
import subprocess as sub
from colorama import Fore, Back, Style
from .utils_datasets import GetDatasets
from .data_h5_storages import HStorage

# ----------------------------------------------------------------
# Command templates
# ----------------------------------------------------------------

def cmd_hdiutil_create(path_to_file, output_file):
    return 'hdiutil create -fs HFS+ -srcfolder \'{path_to_file}/\' -volname \'{filename}\' -verbose \'{output_file}\''.format(path_to_file = path_to_file, filename = os.path.basename(path_to_file), output_file = output_file)

def cmd_hdiutil_mount(path_to_dmg, mountpoint):
    return 'hdiutil attach -mountpoint \'{mountpoint}\' \'{path_to_dmg}\''.format(path_to_dmg = path_to_dmg, mountpoint = mountpoint)

def cmd_hditul_detach(mountpoint):
    return 'hdiutil detach \'{mountpoint}\''.format(mountpoint = mountpoint)

def cmd_delete_files_exclude_dirs(path_to_folder):
    return 'find \'{path_to_folder}\' -maxdepth 1 -type f -delete'.format(path_to_folder = path_to_folder)

def cmd_delete_files(*wildcards):
    '''Return a template for command that deletes all files specify using wildcards arguments list (supports folders)'''
    return 'rm -rfv ' + ' '.join(['\'' + w + '\'' for w in wildcards])

def cmd_copy(source, dest):
    return 'rsync -azvhW --progress \'{source}\' \'{dest}\''.format(source = source, dest = dest)

def cmd_scp(source, dest, port):
    return 'rsync -azvhW --progress --rsh \'ssh -p {port}\' \'{source}\' \'{dest}\''.format(source = source, dest = dest, port = port)

def cmd_move(source, dest):
    return 'mv -v \'{source}\' \'{dest}\''.format(source = source, dest = dest)

def cmd_rename(source, dest):
    return 'mv -v \'{source}\' \'{dest}\''.format(source = source, dest = dest)

def cmd_untar(source, dest):
    return 'tar -xvf \'{source}\' -C \'{dest}\''.format(source = source, dest = dest)


# ----------------------------------------------------------------
# Config Class
# ----------------------------------------------------------------

class BaseTask():
    def __init__(self, config, name = '', context = {}, actions = [], arguments_count = 0, description = ''):
        '''
        Config dictionary is merged with task attributes.
        So you can access the config data just by calling **task.mountpoint** instead of **task.config['mountpoint']**

        If arugments_count > 0, the arguments will be automatically processed and passed as array in the context['arguments'] to the task.
        '''
        self.config = config
        self.actions = actions
        self.context = context
        self.arguments_count = arguments_count
        self.description = description
        self._name = name

        # Save time to context
        from datetime import datetime
        self.init_time = str(datetime.strftime(datetime.now(), '%Y_%m_%d_%H%M%S'))

    @property
    def name(self):
        return self._name

    def setName(self, name, overwrite = True):
        if overwrite:
            self._name = name
        else:
            self._name = self._name or name

    def getDescription(self, name = ''):
        return '{name}: {description}'.format(name = name or self._name, description = self.description)

    def __getattr__(self, attr):
        if attr in self.config:
            return self.config[attr]
        else:
            return None

    def __call__(self, printNonCritical = True):
        prevError = None
        for action in self.actions:
            if action.__name__ is not 'ActionComment':
                print(Style.BRIGHT + Fore.CYAN + 'Action: ' + action.__name__ + '' + Style.RESET_ALL)
                print(Fore.CYAN + '================================================================' + Style.RESET_ALL)

            err = action(self, prevError)

            if err:
                err.method    = action.__name__
                err.prevError = prevError

                if self.ignoreErrors:
                    err.printMe()
                else:
                    if err.isCritical:
                        break
                    else:
                        if printNonCritical:
                            err.printMe()

            prevError = err

            if action.__name__ is not 'ActionComment':
                print(Fore.CYAN + '----------------------------------------------------------------' + Style.RESET_ALL)
        return err

class Task(BaseTask):
    def __init__(self, config, path, **kwargs):
        super().__init__(config, **kwargs)
        self._path = os.path.abspath(os.path.normpath(path))

    @property
    def name(self):
        return self._name or os.path.basename(self.path)

    @property
    def path(self):
        return self._path

    @property
    def mountpoint(self):
        return os.path.join(self.config['mountpoint'], self.name)

    # def getContext(self, contextKey):
    #   return self.context.get(contextKey, '')

    # def setContext(self, contextKey, value):
    #   return self.context[contextKey] = value

class ActionError():
    def __init__(self, message, isCritical = False):
        self.method     = ''
        self.message    = message
        self.isCritical = isCritical
        self.prevError  = None

    def printMe(self):
        typeStyle = Back.YELLOW + Fore.BLACK
        msgStyle  = Fore.YELLOW
        errType   = 'WARNING:'
        if self.isCritical:
            typeStyle = Back.RED
            msgStyle  = Fore.RED
            errType   = 'ERROR'
        print(typeStyle + errType + Style.RESET_ALL + msgStyle + ' [' + self.method + '] - ' + self.message + Style.RESET_ALL)

    def printTrace(self):
        if self.prevError:
            self.prevError.printTrace()

        self.printMe()

class NoError(ActionError):
    def __init__(self):
        super().__init__('', False)

    def printMe(self):
        pass

class FileNotExistsError(ActionError):
    def __init__(self, message = 'File not found.', isCritical = False):
        super().__init__(message, isCritical)

class CommandFailedError(ActionError):
    def __init__(self, message = 'Command failed.', isCritical = False):
        super().__init__(message, isCritical)
    
class ContextIsEmptyError(ActionError):
    def __init__(self, message = 'Context is empty!', isCritical = True):
        super().__init__(message, isCritical)

class ErrorCreator():
    def __init__(self, errorClass = None):
        self.errorClass = errorClass

    def onError(self, message, isCritical = False):
        if self.errorClass:
            return self.errorClass(message, isCritical)
        else:
            return NoError()

# ----------------------------------------------------------------
# Helpers
# ----------------------------------------------------------------

def RunCMD(cmd):
    '''
    Run shell command, you can use predefine command templates as an input argument.
    Returns ErrorCreator - you can get the error by calling onError method with custom message.
    '''
    print(Style.DIM + Fore.BLUE + cmd + Style.RESET_ALL + '\n')
    return ErrorCreator(None if sub.call(cmd, shell = True) is 0 else CommandFailedError)

# def CopyFileLocaly(source, destination):
#   return RunCMD(cmd_copy(source, destination))

# def CopyFileRemote(source, remote_dest, port):
#   return RunCMD(cmd_scp(source, remote_dest, port))

# ----------------------------------------------------------------
# Actions
# - used in task.actions
# ----------------------------------------------------------------

def EmptyAction(task, error):
    return NoError()

def CleanTestDriveFolder(task, error):
    '''Remove all files in test drive folder, but exclude exported CSV files.'''
    if not os.path.exists(task.path):
        return FileNotExistsError('Test Drive folder doesn\'t exists.')
    return RunCMD(cmd_delete_files_exclude_dirs(task.path)).onError('Cannot clean test drive folder!')

def DeleteTestDriveFolder(task, error):
    '''Delete test drive folder'''
    if os.path.exists(task.path):
        return RunCMD(cmd_delete_files(task.path)).onError('Cannot delete Test drive folder.')
    return NoError()

# Action that is changing context
def CreateDMG(task, error):
    '''Create DMG image from given test drive folder.'''
    dmgFile = task.path + '.dmg'
    if not os.path.exists(dmgFile):
        if os.path.exists(task.path):
            task.context['dmg_path'] = os.path.dirname(task.path)
            return RunCMD(cmd_hdiutil_create(task.path, dmgFile)).onError('Cannot create DMG!', isCritical = True)
        else:
            return FileNotExistsError('Test drive folder doesn\'t exists! Cannot create DMG from it.', isCritical = True)
    return NoError()

# Action that is changing context
def CreateDMGAtPath(dmg_path):
    def ActionCreateDMG(task, error):
        '''Create DMG image from given test drive folder.'''
        dmgFile = os.path.join(dmg_path, task.name + '.dmg')
        if not os.path.exists(dmgFile):
            if os.path.exists(task.path):
                task.context['dmg_path'] = dmg_path
                return RunCMD(cmd_hdiutil_create(task.path, dmgFile)).onError('Cannot create DMG!', isCritical = True)
            else:
                return FileNotExistsError('Test drive folder doesn\'t exists! Cannot create DMG from it.', isCritical = True)
        return NoError()
    return ActionCreateDMG

# Action that is using context
def MountDMG(task, error):
    '''Mount created DMG image to desired mountpoint'''
    # Expect existing context
    if 'dmg_path' not in task.context:
        return FileNotExistsError('DMG path is\'t defined in context.')

    dmgFile = os.path.join(task.context['dmg_path'], task.name + '.dmg')

    if not os.path.exists(dmgFile):
        return FileNotExistsError('DMG image (\'{dmgFile}\') doesn\'t exists, cannot mount it.'.format(dmgFile = dmgFile), isCritical = True)

    if os.path.exists(task.mountpoint):
        # No need to mount the image if it's already mounted
        print('DMG is already mounted')
        return NoError()
    return RunCMD(cmd_hdiutil_mount(dmgFile, task.mountpoint)).onError('Cannot mount DMG!', isCritical = True)

def UnmountDMG(task, error):
    '''Detach mounted DMG'''
    if not os.path.exists(task.mountpoint):
        return FileNotExistsError('DMG isn\'t mounted, cannot detach it.')
    return RunCMD(cmd_hditul_detach(task.mountpoint)).onError('Cannot detach DMG!')

def MoveDMGToServer(task, error):
    '''Move DMG image to archive directory'''
    # Expect existing context
    if 'dmg_path' not in task.context:
        return FileNotExistsError('DMG path is\'t defined in context.')

    dmgFile = os.path.join(task.context['dmg_path'], task.name + '.dmg')

    err = RunCMD(cmd_scp(dmgFile, task.archiveServer, task.archiveServerPort)).onError('Cannot copy DMG file. Check if the file exists.')

    if not isinstance(err, NoError):
        return err

    # Test if DMG is successfully archived, if YES remove it
    archivePath = os.path.join(task.archivePath, task.name + '.dmg')
    if os.path.exists(archivePath):
        return RunCMD(cmd_delete_files(dmgFile)).onError('Cannot remove DMG file.')
    else:
        return FileNotExistsError('DMG isn\'t copied!')

def MoveHCacheToServer(task, error):
    '''Move H5 Cache to archive directory'''
    path = os.path.join(task.cachePath, task.name + '.h5')

    err = RunCMD(cmd_scp(path, task.archiveServer, task.archiveServerPort)).onError('Cannot copy H5 cache. Check if the file exists.')

    if not isinstance(err, NoError):
        return err

    # Test if Cache is successfully archived, if YES remove it
    archivePath = os.path.join(task.archivePath, task.name + '.h5')
    if os.path.exists(archivePath):
        return RunCMD(cmd_delete_files(path)).onError('Cannot remove H5 Cache file.')
    else:
        return FileNotExistsError('H5 Cache isn\'t copied!')

def Comment(comment):
    '''Create an Comment Action which prints given comment.'''
    def ActionComment(task, error):
        if not error or isinstance(error, NoError):
            print(Back.CYAN + Fore.BLACK + 'COMMENT:' + Style.RESET_ALL + ' ' + comment + Style.RESET_ALL + '\n')
        return NoError()
    return ActionComment

def PullCSVArchive(task, error):
    '''Copy CSV Archive from server to local path'''
    serverPath = os.path.join(task.archiveServer, task.name + '.tar')
    return RunCMD(cmd_scp(serverPath, os.path.dirname(task.path), task.archiveServerPort)).onError('Cannot copy CSV Archive. Check if the file exists.')

def ExtractCSV(task, error):
    '''Extract CSV files from path.'''
    path = task.path + '.tar'
    if not os.path.exists(path):
        return FileNotExistsError('CSV acrhive doesn\'t exists, cannot extract it.')
    return RunCMD(cmd_untar(path, os.path.dirname(task.path))).onError('Cannot extract TAR.', isCritical = True)

def ExtractCSVFromServer(task, error):
    '''Extract CSV files from archive.'''
    path = os.path.join(task.archivePath, task.name + '.tar')
    if not os.path.exists(path):
        return FileNotExistsError('CSV acrhive doesn\'t exists, cannot extract it.')
    return RunCMD(cmd_untar(path, os.path.dirname(task.path))).onError('Cannot extract TAR.', isCritical = True)

def FixedNameAfterExtract(task, error):
    '''Sometimes the CSV archive contains folder called __csv_archive__ rather then the name of TestDrive, this action check if the extracted file is __csv_archive__ or TestDrive folder.'''
    if not os.path.exists(task.path) or not os.path.isdir(task.path):
        # TestDrive folder doesn't exists, check if __csv_archive__ exists
        archivePath = os.path.join(os.path.dirname(task.path), '__csv_archive__')
        if os.path.exists(archivePath) and os.path.isdir(archivePath):
            # Rename the __csv_archive__ to TestDrive folder
            return RunCMD(cmd_rename(archivePath, task.path)).onError('Cannot rename __csv_archive__.')
        else:
            return FileNotExistsError('TestDrive or __csv_archive__ folder doesn\'t exists, cannot fixed.', isCritical = True)
    # Test Drive folder exists, everything is ok
    return NoError()

def RunCreateDatasets(task, error):
    '''CreateDatasets from mounted DMG image.'''

    ds = GetDatasets(
        path = task.mountpoint,
        storage_class = HStorage,
        cache_path = task.context.get('cache_path', '')
    )
    return NoError() if ds else FileNotExistsError('Dataset cannot be created, because specified path doesn\'t exists.', isCritical = True)

# Repeat action
def Repeat(ntimes, actions):
    def RepeatAction(task, error):
        if isinstance(error, NoError):
            # Create copy of task
            t = Task(config = task.config, path = task.path, actions = [Comment('')] + actions, context = task.context)
            # Need to remember to copy init_time
            t.init_time = task.init_time

            # Run the task t n-times
            for i in range(ntimes):
                t.actions[0] = Comment('Repeating: {} / {}'.format(i + 1, ntimes))
                err = t()
                if err.isCritical:
                    return err
        return error

    return RepeatAction

# Actions changing context
def UpdateContext(**new_context):
    def ActionUpdateContext(task, error):
        task.context.update(new_context)
        return error
    return ActionUpdateContext

def SaveContext(file_path):
    def ActionSaveContext(task, error):
        if isinstance(error, NoError):
            import pickle
            with open(file_path + '.pickle', 'wb') as f:
                pickle.dump(task.context, f, pickle.HIGHEST_PROTOCOL)
        return error
    return ActionSaveContext


def LoadContext(file_path):
    def ActionLoadContext(task, error):
        if not error or isinstance(error, NoError):
            import pickle
            if not os.path.exists(file_path + '.pickle', 'rb'):
                return FileNotExistsError('File \'{}\' doesn\'t exists, cannot load context.'.format(file_path + '.pickle'), isCritical = False)
            with open(file_path + '.pickle') as f:
                context = pickle.load(f)
                if context:
                    task.context.update(context)
                else:
                    return ContextIsEmptyError('Cannot load context, context is empty.', isCritical = False)
        return error
    return ActionLoadContext

# Actions changing config
def UpdateConfig(**new_config):
    def ActionUpdateConfig(task, error):
        task.config.update(new_config)
        return error
    return ActionUpdateConfig

def SaveConfig(file_path):
    def ActionSaveConfig(task, error):
        if isinstance(error, NoError):
            import json
            with open(file_path + '.txt', 'wb') as f:
                f.write(json.dumps(task.config))
        return error
    return ActionSaveConfig

# ----------------------------------------------------------------
# Main Run Tasks
# ----------------------------------------------------------------

def RunTasks(tasks):

    tasks_to_run = []
    if type(tasks) is list:
        tasks_to_run = tasks
        # TODO: Assign default name according to position in the list, if it's not already set. Create dict using {task.name : task}.
    elif type(tasks) is dict:
        # Choose what tasks to run
        import sys
        from collections import deque
        arguments = deque(sys.argv[1:])
        if not arguments:
            tasks_to_run = tasks.values()
        else:
            while len(arguments):
                tname = arguments.popleft()
                # List all availible tasks
                if tname == 'ls':
                    print('\n')
                    print(Fore.MAGENTA + '/' * 64 + Style.RESET_ALL)
                    print(Style.BRIGHT + Fore.MAGENTA + 'Listing tasks in \'{}\':'.format(os.path.basename(sys.argv[0])) + Style.RESET_ALL)
                    print(Fore.MAGENTA + '/' * 64 + Style.RESET_ALL)
                    print(Style.BRIGHT + Fore.WHITE + ' > ' + '\n > '.join(sorted([task.getDescription(tname) for (tname, task) in tasks.items()])) + Style.RESET_ALL)
                    print('\n')
                    return
                assert tname in tasks, 'Task name \'{}\' is not in the tasks list, check the \'{}\' script if it contains all tasks you wanted.'.format(tname, sys.argv[0])

                # Parse arguments if the task expect some
                if tasks[tname].arguments_count:
                    args = []
                    for i in range(tasks[tname].arguments_count):
                        args.append(arguments.pop())
                    tasks[tname].context['arguments'] = args

                # Update task name if it's not already set
                tasks[tname].setName(tname, overwrite = False)
                # Append task to tasks, which will be run
                tasks_to_run.append(tasks[tname])

    # Run all tasks
    for (i, task) in enumerate(tasks_to_run):
        print('\n')
        print(Fore.MAGENTA + '/' * 64 + Style.RESET_ALL)
        print(Style.BRIGHT + Fore.MAGENTA + ('Starting TASK \'{name}\' ({current} from {total}):'.format(name = task.name, current = i+1, total = len(tasks_to_run))) + Style.RESET_ALL)
        print(Fore.MAGENTA + '/' * 64 + Style.RESET_ALL)
        print('\n')
        err = task()
        if err.isCritical:
            print('\n' + Back.RED + 'CRITICAL ERROR FOUND:' + Style.RESET_ALL + ' Printing traceback:' + '\n')
            err.printTrace()

# ----------------------------------------------------------------
# Testing function
# ----------------------------------------------------------------

def Main():
    config = {
        'mountpoint' : '/Volumes/',
        'archivePath' : '/Volumes/Archive/',
        'cachePath' : '/Users/vojtechmicka/Dropbox/Projects/Valeo/Rudy/__h5__'
    }

    tasks = []
    tasks.append(Task(config,
        path = '/Users/vojtechmicka/Dropbox/Projects/Valeo/Test/Data',
        actions = [
            # ExtractCSVArchive,
            # Comment('CSV should be extracted.'),
            CleanTestDriveFolder,
            Comment('Folder is cleaned!'),
            CreateDMG,
            Comment('DMG is created.'),
            DeleteTestDriveFolder,
            Comment('Test drive folder is deleted.'),
            MountDMG,
            Comment('DMG is mounted.'),
            RunCreateDatasets,
            Comment('Datasets is created.'),
            UnmountDMG,
            Comment('DMG is detached.'),
            MoveDMGToServer,
            Comment('DMG should be on the server.')
        ]))

    RunTasks(tasks)

if __name__ == '__main__':
    try:
        Main()
    except KeyboardInterrupt:
        pass
    finally:
        pass