import rudy.utils_logging as log

from rudy.utils_tasks import Task, RunTasks, Comment, UpdateContext, LoadContext, SaveContext, Repeat, NoError, UpdateConfig
from dnn_tasks import LoadDatasets, TrainModel, LoadTrainedModel, EvaluateModel, PredictWithModel, CompileModel
from tasks_models import * # Import all models
from benchmark_datasets import BenchmarkModel

def GetRepeatSubactionsFromDS(datasets):
    repeat_from_ds = []
    for ds in datasets:
        repeat_from_ds.append(LoadDatasets(ds))
        repeat_from_ds.append(Comment('Dataset \'{}\' is loaded.'.format(ds)))
        repeat_from_ds.append(TrainModel)
    return repeat_from_ds

def Main():
    log.addConsoleFormat(level = 'DEBUG')

    batch_size = 128
    img_size = (16, 64, 3)

    config = {
        # What shape the input data have, from the left it is (batch number, time = need to be set to 1, and image size (16, 64, 3))
        'batch_input_shape' : (batch_size, 1) + img_size,
        # Number of epochs the neural net will continue to train
        'epochs' : 100,
        # Path to data or folder __h5__ where the cached is stored.
        'cache_path' : '/media/tesla/PRA1-HDD004',
        # Path where you want to save the experiments (predictions and weights)
        'experiments_path' : '/media/tesla/PRA1-HDD004/Experiments',
        # Moni server address
        'moni_server' : 'http://localhost:9000',
        # If you want to plot training progress
        'plotter' : False,
        # What optimizer you want to run ('adamax', 'sgd')
        'optimizer' : 'adamax',
        # How you want to reset the LSTM ('auto' means at the end of each dataset)
        'lstm_reset' : 'auto',
        # Save weights for each epoch
        'train_checkpoint' : True,
        # Stop early if the neural net will stop learn
        'train_early_stop' : False,
        # Filename for exported weights
        'weights_file' : 'cnn_lstm',
        # Learning rate
        'learn_rate' : 0.1,
        # Fit using generator or load the date into memory (memory intensive, but faster training and you can use shuffle in SimpleForward model)
        'fit_using_generator' : False,
        # Weight for specific classes, 0 ... no alert, 1 ... alert
        'class_weights' : {0: 1, 1: 100},
        # Type of recurrent network you want to use ('lstm', 'gru', 'simple_rnn')
        'rnn_class' : 'lstm',
    }

    datasets = [
        '1_L13040_112603', # few short positive sequences
        '1_L_094631',
        '3_2_5_3PWIAAL2040_155759', # few short positive sequences
        '3_L_151956',
        '3_8_2_1S_095131',# lots of negative frames
        '5_L_155609',
        '1_L32S_144049', # few short positive sequences
        '1_L_143307',
        '4_L_154104', # Weird results, not so much positives, but the model isn't able to learn the data
        '3_1_L2059_153634',
        '3_2_5_2PWIBAL2040_153709',
        '1_L_130249',
        '1_L16S_142427',
        '1_L26575_135518',
        '3_6_4_L_164148',
        #'3_4_2_L60k30k20mSS_104544', # no positive
    ]

    # create_model_action = CreateModelSimpleForward
    create_model_action = CreateModelSimple
    # create_model_action = CreateModelVGG16

    # Specify what tasks you want to run
    tasks = {
        'train_repeat' :\
        Task(config,
                path = 'Func_Test_OTRs',
                actions = [
                    create_model_action,
                    Comment('Model created.'),
                    CompileModel,
                    Comment('Model compiled'),
                    Repeat(2, GetRepeatSubactionsFromDS(datasets)),
                    BenchmarkModel,
                    Comment('Benchmark of model is done.'),
                ]),

        'train_multids' :\
        Task(config,
            path = 'Func_Test_OTRs',
            actions = [
                    create_model_action,
                    CompileModel,
                    Comment('Model created.'),
                    LoadDatasets(datasets),
                    Comment('Multidataset was loaded.'),
                    TrainModel,
                    Comment('Model trained.'),
                    BenchmarkModel,
                    Comment('Benchmark of model is done.'),
                ]),

        'train_fnn_rnn' :\
        Task(config,
            path = 'Func_Test_OTRs',
            actions = [
                # Change batch size for simple FNN
                UpdateConfig(batch_input_shape = (batch_size,) + img_size, shuffle = True),
                CreateModelSimpleForward,
                CompileModel,
                Comment('FNN Model created.'),
                LoadDatasets(datasets),
                Comment('Datasets \'{}\' loaded'.format(datasets)),
                TrainModel,
                Comment('Model trained.'),
                # Change batch size so it contains time dimension and freeze featrue layers
                UpdateConfig(epochs = 200, batch_input_shape = (batch_size, 1) + img_size, freeze_feature_layers = False, load_by_name = True, shuffle = False),
                CreateModelSimple,
                CompileModel,
                Comment('RNN Model created.'),
                LoadTrainedModel('Simple_Forward'),
                Comment('Weights loaded.'),
                LoadDatasets(datasets),
                Comment('Datasets \'{}\' loaded'.format(datasets)),
                TrainModel,
                Comment('Model trained.'),
                # Benchmark model
                BenchmarkModel,
                Comment('Benchmark of model is done.'),
            ]),

        'train_rnn' :\
        Task(config,
            path = 'Func_Test_OTRs',
            actions = [
                # Change batch size so it contains time dimension and freeze featrue layers
                UpdateConfig(batch_input_shape = (batch_size, 1) + img_size, freeze_feature_layers = False, load_by_name = True, shuffle = False),
                CreateModelSimple,
                CompileModel,
                Comment('RNN Model created.'),
                LoadTrainedModel('Simple_Forward'),
                Comment('Weights loaded.'),
                LoadDatasets(datasets),
                Comment('Datasets \'{}\' loaded'.format(datasets)),
                TrainModel,
                Comment('Model trained.'),
                # Benchmark model
                BenchmarkModel,
                Comment('Benchmark of model is done.'),
            ]),

        'train_vgg' :\
        Task(config,
            path = 'Func_Test_OTRs',
            actions = [
                UpdateConfig(batch_input_shape = (batch_size, 1) + img_size),
                CreateModelVGG16,
                CompileModel,
                LoadDatasets(datasets),
                Comment('Datasets \'{}\' loaded'.format(datasets)),
                TrainModel,
                BenchmarkModel,
            ]),

        'benchmark_fnn' :\
        Task(config,
            path = 'Func_Test_OTRs',
            actions = [
                UpdateConfig(batch_input_shape = (batch_size,) + img_size),
                CreateModelSimpleForward,    Comment('Model created.'),
                CompileModel,
                LoadTrainedModel(),     Comment('Trained model loaded.'), # Loaded latest dataset
                BenchmarkModel,         Comment('Benchmark of model is done.'),
            ]),

        'benchmark' :\
        Task(config,
            path = 'Func_Test_OTRs',
            actions = [
                create_model_action,    Comment('Model created.'),
                CompileModel,
                LoadTrainedModel(),     Comment('Trained model loaded.'), # Loaded latest dataset
                BenchmarkModel,         Comment('Benchmark of model is done.'),
            ]),
    }

    RunTasks(tasks)

if __name__ == '__main__':
    try:
        Main()
    except KeyboardInterrupt:
        pass
    finally:
        pass
