import rudy.utils_logging as log

from rudy.utils_tasks import Task, RunTasks, Comment, UpdateContext, LoadContext, SaveContext, Repeat, NoError, UpdateConfig
from dnn_tasks import LoadDatasets, TrainModel, LoadTrainedModel, EvaluateModel, PredictWithModel, FindFilename, CompileModel
from tasks_models import * # Import all models
from benchmark_datasets import PlotBenchmark, BenchmarkModel
from tqdm import tqdm
import numpy as np

def TestDatasets(task, error):
	ds = task.context['dataset']
	print('DS.choosen_side', ds.choosen_side, type(ds.choosen_side))
	print('DS.fixed_beams', ds.fixed_beams, type(ds.fixed_beams))
	print('DS.sizes', ds.sizes, type(ds.sizes), type(ds.sizes[0]), type(ds.sizes[1]))

	for name, _ds in zip(['TRAIN', 'TEST', 'ALL'], [ds.train, ds.test, ds.all]):
		print('Testing:', name)
		gen = _ds.generator
		print('Size:', _ds.size)
		pos = 0
		for i in tqdm(range(_ds.size)):
			t = next(gen)
			pos += np.sum(t[1])
		print(i, t[0].shape, t[1].shape, t[1].dtype, 'pos:', pos)
	return NoError()

def TestFindFilename(task, error):
	print('SAVING FILE:')
	save = FindFilename(root = task.experiments_path, subfolder = 'Weights', name = '{}_{}.h5'.format(task.path, task.context['batch_name']), timefolder = task.init_time)
	print(save)

	print('LOADING FILE:')
	print(FindFilename(root = task.experiments_path, subfolder = 'Weights', name = '{}_{}.h5'.format(task.path, task.context['batch_name']), find_last = True, use_timefolder = True))

	import os
	os.makedirs(os.path.dirname(save))
	return NoError()

def Main():
	log.addConsoleFormat(level = 'DEBUG')

	batch_size = 32
	img_size = (16, 64, 3)

	config = {
		# What shape the input data have, from the left it is (batch number, time = need to be set to 1, and image size (16, 64, 3))
		'batch_input_shape' : (batch_size, 1) + img_size,
		# Number of epochs the neural net will continue to train
        'epochs' : 1,
        # Path to data or folder __h5__ where the cached is stored.
        'cache_path' : '.',
        # Path where you want to save the experiments (predictions and weights)
        'experiments_path' : '../Predictions/',
        # Moni server address
        # 'moni_server' : 'http://localhost:9000',
        # If you want to plot training progress
        'plotter' : False,
        # What optimizer you want to run ('adamax', 'sgd')
        'optimizer' : 'adamax',
        # How you want to reset the LSTM ('auto' means at the end of each dataset)
        'lstm_reset' : 'auto',
        # Save weights for each epoch
        'train_checkpoint' : True,
        # Stop early if the neural net will stop learn
        'train_early_stop' : True,
        # Filename for exported weights
        'weights_file' : 'cnn_lstm',
        # Learning rate
        'learn_rate' : 0.1,
        # Fit using generator or load the date into memory (memory intensive, but faster training and you can use shuffle in SimpleForward model)
        'fit_using_generator' : False,
        # Weight for specific classes, 0 ... no alert, 1 ... alert
        'class_weights' : {0: 1, 1: 100},
        # Type of recurrent network you want to use ('lstm', 'gru', 'simple_rnn')
        'rnn_class' : 'lstm',
	}

	datasets = '1_L13040_112603'

	# Specify what tasks you want to run
	tasks = {
		'test_dnn' :\
		Task(config,
			path = 'Test',
			actions = [
				CreateModelSimple,					Comment('Model created.'),
				CompileModel,
				LoadDatasets('3_8_2_1S_095131'),	Comment('Dataset \'3_8_2_1S_095131\' loaded.'),
				TrainModel,
			]),

		'benchmark' :\
		Task({**config, 'weights_file' : 'Func_Test_OTRs_1_L13040_112603_weights'},
			path = 'Test',
			actions = [
				CreateModelSimple,					Comment('Model created.'),
				CompileModel,
				LoadTrainedModel(),					Comment('Trained model loaded.'),
				BenchmarkModel,						Comment('Benchmark of model is done.'),
			]),

		'test_ds' :\
		Task(config,
			path = 'Test',
			actions = [
				LoadDatasets('3_8_2_1S_095131'), Comment('Single dataset is loaded.'),
				TestDatasets,
				LoadDatasets(['1_L13040_112603', '3_8_2_1S_095131']), Comment('Multiple dataset is loaded.'),
				TestDatasets,
			]),

		'test_multids' :\
		Task(config,
			path = 'Test',
			actions = [
					CreateModelSimple,
					CompileModel,
					Comment('Model created.'),
					LoadDatasets([
						'1_L13040_112603',
						'3_8_2_1S_095131',
					]),
					Comment('Multidataset was loaded.'),
		            TrainModel,
		            Comment('Model trained.'),
					BenchmarkModel,
					Comment('Benchmark of model is done.'),
				]),

		'show_benchmark' :\
		Task(config,
			path = 'Test',
			actions = [
				# PlotBenchmark('../Predictions/2016_09_29_182320_subset/Benchmarks.Simple_Forward', limit = 0, skip_plot = False),
				PlotBenchmark('../Predictions/2016_09_29_182320_subset/Benchmarks.Simple_LSTM', limit = 0, skip_plot = False),
				PlotBenchmark('../Predictions/2016_09_30_111624_subset/Benchmarks.VGG16', limit = 0, skip_plot = False),
				# PlotBenchmark('../Predictions/2016_09_27_015314/Benchmarks', limit = 0, just_all = True),
				# PlotBenchmark('../Predictions/2016_09_25_141654/Benchmarks', limit = 0, just_all = True),
			]),

		'test_findfile' :\
		Task(config,
			path = 'Test',
			actions = [
				LoadDatasets('1_L13040_112603'),
				TestFindFilename,
			]),

		'test_createmodel_rnn' :\
		Task(config,
			path = 'Test',
			actions = [
				CreateModelSimple
			]),

		'test_createmodel_fnn' :\
		Task({**config, 'batch_input_shape' : (batch_size,) + img_size},
			path = 'Test',
			actions = [
				CreateModelSimpleForward
			]),

		'show_models_summary' :\
		Task(config,
			path = 'Test',
			actions = [
				CreateModelSimple,

				UpdateConfig(batch_input_shape = (batch_size,) + img_size, shuffle = True),
				CreateModelSimpleForward,

				UpdateConfig(batch_input_shape = (batch_size, 1) + img_size),
                CreateModelVGG16,
			]),

		'train_fnn_rnn' :\
		Task(config,
			path = 'Test',
			actions = [
				# Change batch size for simple FNN
				UpdateConfig(batch_input_shape = (batch_size,) + img_size, shuffle = True),
				CreateModelSimpleForward,
				CompileModel,
				Comment('FNN Model created.'),
				LoadDatasets(datasets),
				Comment('Datasets \'{}\' loaded'.format(datasets)),
				TrainModel,
				Comment('Model trained.'),
				# Change batch size so it contains time dimension and freeze featrue layers
				UpdateConfig(batch_input_shape = (batch_size, 1) + img_size, freeze_feature_layers = True, load_by_name = True, shuffle = False),
				CreateModelSimple,
				CompileModel,
				Comment('RNN Model created.'),
				LoadTrainedModel('Simple_Forward'),
				Comment('Weights loaded.'),
				LoadDatasets(datasets),
				Comment('Datasets \'{}\' loaded'.format(datasets)),
				TrainModel,
				Comment('Model trained.'),
				# Benchmark model
				BenchmarkModel,
				Comment('Benchmark of model is done.'),
			]),

        'train_vgg' :\
        Task(config,
            path = 'Test',
            actions = [
                UpdateConfig(batch_input_shape = (batch_size, 1) + img_size),
                CreateModelVGG16,
                CompileModel,
                LoadDatasets(datasets),
                Comment('Datasets \'{}\' loaded'.format(datasets)),
                TrainModel,
                BenchmarkModel,
            ]),
	}

	RunTasks(tasks)

if __name__ == '__main__':
    try:
        Main()
    except KeyboardInterrupt:
        pass
    finally:
        pass
