from rudy.utils_tasks import NoError

from keras.models import Sequential
from keras.layers.core import Dense, Activation, Flatten
from keras.layers.normalization import BatchNormalization
from keras.layers.convolutional import Convolution2D, MaxPooling2D
from keras.layers.recurrent import LSTM, GRU, SimpleRNN
from keras.layers.wrappers import TimeDistributed

def CreateModelSimpleForward(task, error):
    '''Action that creates simple model with FNN as classification.'''

    feature_layers = [
        Convolution2D(32, 3, 3, border_mode='same', batch_input_shape = task.batch_input_shape, name = 'conv1_1'),
        Activation('relu', name = 'active1_1'),
        MaxPooling2D(pool_size = (2,2), name = 'maxpool1_1'),
        BatchNormalization(name = 'bn1_1'),

        Convolution2D(32, 3, 3, border_mode='same', name = 'conv2_1'),
        Activation('relu', name = 'active2_1'),
        MaxPooling2D(pool_size = (2,2), name = 'maxpool2_1'),
        BatchNormalization(name = 'bn2_1'),

        Convolution2D(32, 3, 3, border_mode='same', name = 'conv3_1'),
        Activation('relu', name = 'active3_1'),
        MaxPooling2D(pool_size = (2,2), name = 'maxpool3_1'),
        BatchNormalization(name = 'bn3_1'),

        Convolution2D(32, 2, 2, border_mode='same', name = 'conv4_1'),
        Activation('relu', name = 'active4_1'),
        MaxPooling2D(pool_size = (2,2), name = 'maxpool4_1'),
        BatchNormalization(name = 'bn4_1'),

        Flatten(name = 'flat5_1'),
        # Activation('relu', name = 'active5_1'),
        # BatchNormalization(name = 'bn5_1'),
    ]

    classifications_layers = [
        Dense(32, name = 'f_dense1_1'),
        Activation('relu', name = 'f_active1_1'),
        BatchNormalization(name = 'f_bn1_1'),

        Dense(1, name = 'f_dense2_1'),
        Activation('sigmoid', name = 'f_active2_1'),
    ]

    # Freeze feature layers
    if task.freeze_feature_layers:
        for l in feature_layers:
            l.trainable = False

    # Create simple model
    model = Sequential(feature_layers + classifications_layers)

    # Print model summary
    model.summary()

    task.context['model'] = model
    task.context['model_id'] = 'Simple_Forward'

    return NoError()

def CreateModelSimple(task, error):
    '''Action that simple model with LSTM layers'''

    feature_layers = [
        TimeDistributed(Convolution2D(32, 3, 3, border_mode='same'), batch_input_shape = task.batch_input_shape, name = 'conv1_1'),
        TimeDistributed(Activation('relu'), name = 'active1_1'),
        TimeDistributed(MaxPooling2D(pool_size = (2,2)), name = 'maxpool1_1'),
        TimeDistributed(BatchNormalization(), name = 'bn1_1'),

        TimeDistributed(Convolution2D(32, 3, 3, border_mode='same'), name = 'conv2_1'),
        TimeDistributed(Activation('relu'), name = 'active2_1'),
        TimeDistributed(MaxPooling2D(pool_size = (2,2)), name = 'maxpool2_1'),
        TimeDistributed(BatchNormalization(), name = 'bn2_1'),

        TimeDistributed(Convolution2D(32, 3, 3, border_mode='same'), name = 'conv3_1'),
        TimeDistributed(Activation('relu'), name = 'active3_1'),
        TimeDistributed(MaxPooling2D(pool_size = (2,2)), name = 'maxpool3_1'),
        TimeDistributed(BatchNormalization(), name = 'bn3_1'),

        TimeDistributed(Convolution2D(32, 2, 2, border_mode='same'), name = 'conv4_1'),
        TimeDistributed(Activation('relu'), name = 'active4_1'),
        TimeDistributed(MaxPooling2D(pool_size = (2,2)), name = 'maxpool4_1'),
        TimeDistributed(BatchNormalization(), name = 'bn4_1'),

        TimeDistributed(Flatten(), name = 'flat5_1'),
        # TimeDistributed(Activation('relu'), name = 'active5_1'),
        # TimeDistributed(BatchNormalization(), name = 'bn5_1'),
    ]

    reccurent_layers = []
    if task.rnn_class == 'lstm':
        reccurent_layers = [
            LSTM(128, stateful = True, return_sequences = True, name = 'r_lstm1_1'),
            LSTM(128, stateful = True, return_sequences = False, name = 'r_lstm1_2'),
        ]
    elif task.rnn_class == 'gru':
        reccurent_layers = [
            GRU(128, stateful = True, return_sequences = False, name = 'r_gru1_1'),
            # GRU(128, stateful = True, return_sequences = False, name = 'r_gru1_2'),
        ]
    else: # 'simple_rnn'
        reccurent_layers = [
            SimpleRNN(128, stateful = True, return_sequences = False, name = 'r_rnn1_1'),
            # SimpleRNN(128, stateful = True, return_sequences = False, name = 'r_rnn1_2'),
        ]
    # elif task.rnn_class == 'simple_rnn':
    #     pass

    classifications_layers = reccurent_layers + [
        Dense(1, name = 'r_dense2_1'),
        Activation('sigmoid', name = 'r_active2_1'),
    ]

    # Freeze feature layers
    if task.freeze_feature_layers:
        for l in feature_layers:
            l.trainable = False

    # Create simple model
    model = Sequential(feature_layers + classifications_layers)

    # Print model summary
    model.summary()

    task.context['model'] = model
    task.context['model_id'] = 'Simple_{}'.format(task.rnn_class.upper() if task.rnn_class else 'RNN')

    return NoError()

def CreateModelVGG16(task, error):
    '''Action that creates and compile model'''
    # Create simple model
    model = Sequential()

    model.add(TimeDistributed(Convolution2D(32, 3, 3, border_mode='same'), batch_input_shape = task.batch_input_shape))
    model.add(TimeDistributed(Activation('relu')))
    model.add(TimeDistributed(BatchNormalization()))
    model.add(TimeDistributed(Convolution2D(32, 3, 3, border_mode='same')))
    model.add(TimeDistributed(Activation('relu')))
    model.add(TimeDistributed(MaxPooling2D(pool_size = (2,2))))
    model.add(TimeDistributed(BatchNormalization()))

    model.add(TimeDistributed(Convolution2D(64, 3, 3, border_mode='same')))
    model.add(TimeDistributed(Activation('relu')))
    model.add(TimeDistributed(BatchNormalization()))
    model.add(TimeDistributed(Convolution2D(64, 3, 3, border_mode='same')))
    model.add(TimeDistributed(Activation('relu')))
    model.add(TimeDistributed(MaxPooling2D(pool_size = (2,2))))
    model.add(TimeDistributed(BatchNormalization()))

    model.add(TimeDistributed(Convolution2D(128, 3, 3, border_mode='same')))
    model.add(TimeDistributed(Activation('relu')))
    model.add(TimeDistributed(BatchNormalization()))
    model.add(TimeDistributed(Convolution2D(128, 3, 3, border_mode='same')))
    model.add(TimeDistributed(Activation('relu')))
    model.add(TimeDistributed(MaxPooling2D(pool_size = (2,2))))
    model.add(TimeDistributed(BatchNormalization()))

    model.add(TimeDistributed(Convolution2D(128, 2, 2, border_mode='same')))
    model.add(TimeDistributed(Activation('relu')))
    model.add(TimeDistributed(BatchNormalization()))
    model.add(TimeDistributed(Convolution2D(128, 2, 2, border_mode='same')))
    model.add(TimeDistributed(Activation('relu')))
    model.add(TimeDistributed(MaxPooling2D(pool_size = (2,2))))
    model.add(TimeDistributed(BatchNormalization()))

    model.add(TimeDistributed(Flatten()))

    model.add(LSTM(64, stateful = True, return_sequences = True))
    model.add(LSTM(64, stateful = True, return_sequences = False))

    model.add(Dense(1))
    model.add(Activation('sigmoid'))

    # Print model summary
    model.summary()

    task.context['model'] = model
    task.context['model_id'] = 'VGG16'

    return NoError()
